"""PixelFly library bindings (:mod:`pymanip.video.pco.pixelfly`)
================================================================

This module implements bindings to the PCO PixelFly library using :mod:`ctypes`.
Please not that these bindings are not official, and that not all PixelFly functions
are wrapped. Please refer to the official PCO PixelFly documentation for accurate description
of the functions.

.. autoclass:: PCO_Image
   :members:
   :private-members:

.. autofunction:: PCO_manage_error

.. autofunction:: bcd_byte_to_str

.. autofunction:: bcd_to_int

.. autofunction:: PCO_OpenCamera

.. autofunction:: PCO_OpenCameraEx

.. autofunction:: PCO_CloseCamera

.. autofunction:: PCO_GetInfoString

.. autofunction:: PCO_GetROI

.. autofunction:: PCO_SetROI

.. autofunction:: PCO_GetFrameRate

.. autofunction:: PCO_SetFrameRate

.. autofunction:: PCO_GetCameraName

.. autofunction:: PCO_GetGeneral

.. autofunction:: PCO_GetSensorStruct

.. autofunction:: PCO_GetCameraDescription

.. autofunction:: PCO_GetCameraHealthStatus

.. autofunction:: PCO_GetRecordingStruct

.. autofunction:: PCO_GetSizes

.. autofunction:: PCO_AllocateBuffer

.. autofunction:: PCO_FreeBuffer

.. autofunction:: PCO_GetBufferStatus

.. autofunction:: PCO_ArmCamera

.. autofunction:: PCO_GetRecordingState

.. autofunction:: PCO_SetRecordingState

.. autofunction:: PCO_GetBitAlignment

.. autofunction:: PCO_SetBitAlignment

.. autofunction:: PCO_GetImageStruct

.. autofunction:: PCO_GetMetaData

.. autofunction:: PCO_SetMetaDataMode

.. autofunction:: PCO_GetMetaDataMode

.. autofunction:: PCO_SetTimestampMode

.. autofunction:: PCO_AddBufferEx

.. autofunction:: PCO_CancelImages

.. autofunction:: PCO_SetImageParameters

.. autofunction:: PCO_GetImageEx

.. autofunction:: PCO_SetDelayExposureTime

.. autofunction:: PCO_GetDelayExposureTime

.. autofunction:: PCO_GetTriggerMode

.. autofunction:: PCO_SetTriggerMode

.. autofunction:: PCO_SetADCOperation

.. autofunction:: PCO_GetADCOperation

.. autofunction:: PCO_SetPixelRate

.. autofunction:: PCO_GetPixelRate

.. autofunction:: PCO_GetNoiseFilterMode

.. autofunction:: PCO_SetNoiseFilterMode

.. data:: PCO_TriggerModeDescription

   dictionnary of trigger modes

"""

from pathlib import Path
import ctypes

try:
    from ctypes.wintypes import WORD, DWORD, BYTE, SHORT, LONG, HANDLE
except ValueError:
    # ValueError is raised when sphinx tries to import it on Mac and Linux
    WORD = ctypes.c_int
    DWORD = ctypes.c_longlong
    BYTE = ctypes.c_char
    SHORT = ctypes.c_short
    LONG = ctypes.c_long
    HANDLE = ctypes.c_void_p

from enum import IntEnum
from typing import Tuple, Iterable
import datetime

# Open DLL
pixelfly_possible_dllpath = [
    Path(r"C:\Program Files\Digital Camera Toolbox\Camware4\SC2_Cam.dll"),
    Path(r"C:\Program Files\PCO Digital Camera Toolbox\pco.camware\SC2_Cam.dll"),
]
for pixelfly_dllpath in pixelfly_possible_dllpath:
    if pixelfly_dllpath.exists():
        pixelfly_dll = ctypes.windll.LoadLibrary(str(pixelfly_dllpath))
        break
else:
    print(pixelfly_dllpath, "not found")


# General constants
class PCO_ErrorLayer(IntEnum):
    PCO_ERROR_FIRMWARE = 0x00001000  # error inside the firmware
    PCO_ERROR_DRIVER = 0x00002000  # error inside the driver
    PCO_ERROR_SDKDLL = 0x00003000  # error inside the SDK-dll
    PCO_ERROR_APPLICATION = 0x00004000  # error inside the application


class PCO_ErrorWarningSource(IntEnum):
    SC2_ERROR_PCOCAM_POWER_CPLD = 0x00010000  # error at CPLD in pco.power unit
    SC2_ERROR_PCOCAM_HEAD_UP = 0x00020000  # error at uP of head board in pco.camera
    SC2_ERROR_PCOCAM_MAIN_UP = 0x00030000  # error at uP of main board in pco.camera
    SC2_ERROR_PCOCAM_FWIRE_UP = (
        0x00040000  # error at uP of firewire board in pco.camera
    )
    SC2_ERROR_PCOCAM_MAIN_FPGA = 0x00050000  # error at FPGA of main board in pco.camera
    SC2_ERROR_PCOCAM_HEAD_FPGA = 0x00060000  # error at FGPA of head board in pco.camera
    SC2_ERROR_PCOCAM_MAIN_BOARD = 0x00070000  # error at main board in pco.camera
    SC2_ERROR_PCOCAM_HEAD_CPLD = 0x00080000  # error at CPLD of head board in pco.camera
    SC2_ERROR_SENSOR = 0x00090000  # error at image sensor (CCD or CMOS)
    SC2_ERROR_SDKDLL = 0x000A0000  # error inside the SDKDLL
    SC2_ERROR_DRIVER = 0x000B0000  # error inside the driver
    SC2_ERROR_POWER = 0x000D0000  # error within power unit
    PCO_ERROR_CAMWARE = 0x00100000  # error in CamWare (also some kind of "device")
    PCO_ERROR_CONVERTDLL = 0x00110000  # error inside the convert dll


# The following bits values are those returned by
# PCO_GetCameraHealthStatus
# cf pco.sdk_manual page 36, 37, 38


class PCO_WarningBits(IntEnum):
    WARNING_POWER_SUPPLY_VOLTAGE = 0x00000001
    WARNING_POWER_SUPPLY_TEMPERATURE = 0x00000002
    WARNING_CAMERA_TEMPERATURE = 0x00000004
    WARNING_SENSOR_TEMPERATURE = 0x00000008
    WARNING_BATTERY_DISCHARGED = 0x00000010
    WARNING_OFFSET_REGUL_RANGE = 0x00000020


class PCO_ErrorBits(IntEnum):
    ERROR_POWER_SUPPLY_VOLTAGE = 0x00000001
    ERROR_POWER_SUPPLY_TEMPERATURE = 0x00000002
    ERROR_CAMERA_TEMPERATURE = 0x00000004
    ERROR_SENSOR_TEMPERATURE = 0x00000008
    ERROR_BATTERY_DISCHARGED = 0x00000010
    ERROR_CAMERA_INTERFACE_FAILURE = 0x00010000
    ERROR_CAMERA_RAM_FAILURE = 0x00020000
    ERROR_CAMERA_MAIN_BOARD_FAILURE = 0x00040000
    ERROR_CAMERA_HEAD_BOARD_FAILURE = 0x00080000


class PCO_StatusBits(IntEnum):
    STATUS_DEFAULT_STATE = 0x00000001
    STATUS_SETTINGS_VALID = 0x00000002
    STATUS_RECORDING_STATE = 0x00000004
    STATUS_SENSOR_READOUT_STATE = 0x00000008
    STATUS_FRAMERATE_STATE = 0x00000010
    STATUS_TRIGGERED_STOP = 0x00000020
    STATUS_CAMERA_LOCKED_TO_EXT = 0x00000040
    STATUS_BATTERY_CONNECTED = 0x00000080
    STATUS_POWER_SAVE = 0x00000100
    STATUS_POWER_SAVE_LEFT = 0x00000200
    STATUS_CAMERA_LOCKED_TO_IRIG = 0x00000400
    STATUS_RESERVED = 0x80000000


# Error handling


class PCO_Error(Exception):
    pass


class PCO_Warning(RuntimeWarning):
    pass


def PCO_manage_error(ret_code):
    """This function raises an error exception or a runtime warning if ret_code is non-zero.

    :param ret_code: PCO library function return code
    :type ret_code: int
    """
    if ret_code != 0:
        f = pixelfly_dll.PCO_GetErrorText
        f.argtypes = (DWORD, ctypes.c_char_p, DWORD)
        desc = ctypes.create_string_buffer(256)
        f(ret_code, desc, 256)
        description = desc.raw.decode("ascii")
        if "warning" in description:
            raise PCO_Warning(description)
        else:
            raise PCO_Error(description)


# C Structure definitions
class PCO_SC2_Hardware_DESC(ctypes.Structure):
    _fields_ = [
        ("szName", ctypes.c_char * 16),
        ("wBatchNo", WORD),
        ("wRevision", WORD),
        ("wVariant", WORD),
        ("ZZwDummy", WORD * 20),
    ]


PCO_MAXVERSIONHW = 10


class PCO_HW_Vers(ctypes.Structure):
    _fields_ = [("BoardNum", WORD), ("Board", PCO_SC2_Hardware_DESC * PCO_MAXVERSIONHW)]


class PCO_SC2_Firmware_DESC(ctypes.Structure):
    _fields_ = [
        ("szName", ctypes.c_char * 16),
        ("bMinorRev", BYTE),
        ("bMajorRev", BYTE),
        ("wVariant", WORD),
        ("ZZwDummy", WORD * 22),
    ]


PCO_MAXVERSIONFW = 10


class PCO_FW_Vers(ctypes.Structure):
    _fields_ = [
        ("DeviceNum", WORD),
        ("Device", PCO_SC2_Firmware_DESC * PCO_MAXVERSIONFW),
    ]


class PCO_CameraType(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("wCamType", WORD),
        ("wCamSubType", WORD),
        ("ZZwAlignDummy1", WORD),
        ("dwSerialNumber", DWORD),
        ("dwHWVersion", DWORD),
        ("dwFWVersion", DWORD),
        ("wInterfaceType", WORD),
        ("strHardwareVersion", PCO_HW_Vers),
        ("strFirmwareVersion", PCO_FW_Vers),
        ("ZZwDummy", WORD * 39),
    ]

    def __init__(self):
        super(PCO_CameraType, self).__init__()
        self.wSize = ctypes.sizeof(PCO_CameraType)

    def __str__(self):
        return """CamType: {type:}
CamSubType: {subtype:}
Serial number: {serial:}
HW Version: {hw:}
FW Version: {fw:}
Interface type: {inter:}""".format(
            type=self.wCamType,
            subtype=self.wCamSubType,
            serial=self.dwSerialNumber,
            hw=self.dwHWVersion,
            fw=self.dwFWVersion,
            inter=self.wInterfaceType,
        )


class PCO_General(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("ZZwAlignDummy1", WORD),
        ("strCamType", PCO_CameraType),
        ("dwCamHealthWarnings", DWORD),
        ("dwCamHealthErrors", DWORD),
        ("dwCamHealthStatus", DWORD),
        ("sCCDTemperature", SHORT),
        ("sCamTemperature", SHORT),
        ("sPowerSupplyTemperature", SHORT),
        ("ZZwDummy", WORD * 37),
    ]

    def __init__(self):
        super(PCO_General, self).__init__()
        self.wSize = ctypes.sizeof(PCO_General)
        self.strCamType.__init__()

    def __str__(self):
        return """CamType = [{camType:}]
Health Warnings = {warn:}
Health Errors = {err:}
Health Status = {stat:}
CCD Temperature = {ccd:}
Power Supply Temperature = {power:}""".format(
            camType=str(self.strCamType),
            warn=self.dwCamHealthWarnings,
            err=self.dwCamHealthErrors,
            stat=self.dwCamHealthStatus,
            ccd=self.sCCDTemperature,
            power=self.sPowerSupplyTemperature,
        )


PCO_RECORDINGDUMMY = 22


class PCO_Recording(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("wStorageMode", WORD),
        ("wRecSubmode", WORD),
        ("wRecState", WORD),
        ("wAcquMode", WORD),
        ("wAcquEnableStatus", WORD),
        ("ucDay", BYTE),
        ("ucMonth", BYTE),
        ("wYear", WORD),
        ("wHour", WORD),
        ("ucMin", BYTE),
        ("ucSec", BYTE),
        ("wTimeSampMode", WORD),
        ("wRecordStopEventMode", WORD),
        ("dwRecordStopDelayImages", DWORD),
        ("wMetaDataMode", WORD),
        ("wMetaDataSize", WORD),
        ("wMetaDataVersion", WORD),
        ("ZZwDummy1", WORD),
        ("dwAcquModeExNumberImages", DWORD),
        ("dwAcquModeExReserved", DWORD * 4),
        ("ZZwDummy", WORD * PCO_RECORDINGDUMMY),
    ]

    def __init__(self):
        super(PCO_Recording, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Recording)


class PCO_Description(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("wSensorTypeDESC", WORD),
        ("wSensorSubTypeDESC", WORD),
        ("wMaxHorzResStdDESC", WORD),
        ("wMaxVertResStdDESC", WORD),
        ("wMaxHorzResExtDESC", WORD),
        ("wMaxVertResExtDESC", WORD),
        ("wDynResDESC", WORD),
        ("wMaxBinHorzDESC", WORD),
        ("wBinHorzSteppingDESC", WORD),
        ("wMaxBinVertDESC", WORD),
        ("wBinVertSteppingDESC", WORD),
        ("wRoiHorStepsDESC", WORD),
        ("wRoiVertStepsDESC", WORD),
        ("wNumADCsDESC", WORD),
        ("wMinSizeHorzDESC", WORD),
        ("dwPixelRateDESC", DWORD * 4),
        ("ZZdwDummypr", DWORD * 20),
        ("wConvFactDESC", WORD * 4),
        ("sCoolingSetPoints", SHORT * 10),
        ("ZZdwDummycv", WORD * 8),
        ("wSoftRoiHozStepsDESC", WORD),
        ("wSoftRoiVertStepsDESC", WORD),
        ("wIRDESC", WORD),
        ("wMinSizeVertDESC", WORD),
        ("dwMinDelayDESC", DWORD),
        ("dwMaxDelayDESC", DWORD),
        ("dwMinDelayStepDESC", DWORD),
        ("dwMinExposureDESC", DWORD),
        ("dwMaxExposureDESC", DWORD),
        ("dwMinExposureStepDESC", DWORD),
        ("dwMinDelayIRDESC", DWORD),
        ("dwMaxDelayIRDESC", DWORD),
        ("dwMinExposureIRDESC", DWORD),
        ("dwMaxExposureIRDESC", DWORD),
        ("wTimeTableDESC", WORD),
        ("wDoubleImageDESC", WORD),
        ("sMinCoolSetDESC", SHORT),
        ("sMaxCoolSetDESC", SHORT),
        ("sDefaultCoolSetDESC", SHORT),
        ("wPowerDownModeDESC", WORD),
        ("wOffsetRegulationDESC", WORD),
        ("wColorPatternDESC", WORD),
        ("wPatternTypeDESC", WORD),
        ("wDummy1", WORD),
        ("wDummy2", WORD),
        ("wNumCoolingSetpoints", WORD),
        ("dwGeneralCapsDESC1", DWORD),
        ("dwGeneralCapsDESC2", DWORD),
        ("dwExtSyncFrequency", DWORD * 4),
        ("dwGeneralCapsDESC3", DWORD),
        ("dwGeneralCapsDESC4", DWORD),
        ("ZZdwDummy", DWORD * 40),
    ]

    def __init__(self):
        super(PCO_Description, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Description)

    @property
    def maximum_resolution_std(self):
        """
        Maximum (horz, vert) resolution in std mode
        """
        return (self.wMaxHorzResStdDESC, self.wMaxVertResStdDESC)

    @property
    def maximum_resolution_ext(self):
        """
        Maximum (horz, vert) resolution in ext mode
        """
        return (self.wMaxHorzResExtDESC, self.wMaxVertResExtDESC)

    @property
    def dynamic_resolution(self):
        """
        Dynamic resolution of ADC in bit
        """
        return self.wDynResDESC

    @property
    def possible_pixelrate(self):
        """
        Possible pixelrate in Hz
        """
        return tuple(self.dwPixelRateDESC)

    @property
    def possible_delay(self):
        """
        Min delay (ns), Max delay (ms), Step (ns)
        """
        return (self.dwMinDelayDESC, self.dwMaxDelayDESC, self.dwMinDelayStepDESC)

    @property
    def possible_exposure(self):
        """
        Min exposure (ns), Max exposure (ms), Step (ns)
        """
        return (
            self.dwMinExposureDESC,
            self.dwMaxExposureDESC,
            self.dwMinExposureStepDESC,
        )

    def nth_cap(self, n):
        return (self.dwGeneralCapsDESC1 >> n) & 0x0001 == 0x0001

    @property
    def general_capabilities(self):
        return {
            "Noisefilter available": self.nth_cap(0),
            "Hotpixelfilter available": self.nth_cap(1),
            "Hotpixel works only with noisefilter": self.nth_cap(2),
            "Timestamp ASCII only available (Timestamp mode 3 enabled)": self.nth_cap(
                3
            ),
            "Dataformat 2x12": self.nth_cap(4),
            "Record Stop Event available": self.nth_cap(5),
            "Hot Pixel correction": self.nth_cap(6),
            "Ext.Exp.Ctrl. not available": self.nth_cap(7),
            "Timestamp not available": self.nth_cap(8),
            "Acquire mode not available": self.nth_cap(9),
            "Dataformat 4x16": self.nth_cap(10),
            "Dataformat 5x16": self.nth_cap(11),
            "Camera has no internal recorder memory": self.nth_cap(12),
            "Camera can be set to fast timing mode (PIV)": self.nth_cap(13),
            "Camera can produce metadata": self.nth_cap(14),
            "Camera allows Set/GetFrameRate cmd": self.nth_cap(15),
            "Camera has Correlated Double Image Mode": self.nth_cap(16),
            "Camera has CCM": self.nth_cap(17),
            "Camera can be synched externally": self.nth_cap(18),
            "Global shutter setting not available": self.nth_cap(19),
            "Camera supports global reset rolling readout": self.nth_cap(20),
            "Camera supports extended acquire command": self.nth_cap(21),
            "Camera supports fan control command": self.nth_cap(22),
            "Camera vert.ROI must be symmetrical to horizontal axis": self.nth_cap(23),
            "Camera horz.ROI must be symmetrical to vertical axis": self.nth_cap(24),
            "Camera has cooling setpoints instead of cooling range": self.nth_cap(25),
            "HW_IO_SIGNAL_DESCRIPTOR available": self.nth_cap(30),
            "Enhanced descriptor available": self.nth_cap(31),
        }

    def __str__(self):
        desc = """Maximum resolution (STD): {:} x {:}
Maximum resolution (EXT): {:} x {:}
Dynamic resolution: {:} bits
Possible pixel rates: {:} Hz
Possible delay: min {:} ns, max {:} ms, step {:} ns
Possible exposure: min {:} ns, max {:} ms, step {:} ns""".format(
            self.maximum_resolution_std[0],
            self.maximum_resolution_std[1],
            self.maximum_resolution_ext[0],
            self.maximum_resolution_ext[1],
            self.dynamic_resolution,
            str(self.possible_pixelrate),
            self.possible_delay[0],
            self.possible_delay[1],
            self.possible_delay[2],
            self.possible_exposure[0],
            self.possible_exposure[1],
            self.possible_exposure[2],
        )
        caps_dict = self.general_capabilities
        caps = "\n".join([k + ":" + str(caps_dict[k]) for k in caps_dict])
        return desc + "\n" + caps


class PCO_Description2(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("ZZwAlignDummy1", WORD),
        ("dwMinPeriodicalTimeDESC2", DWORD),
        ("dwMaxPeriodicalTimeDESC2", DWORD),
        ("dwMinPeriodicalConditionDESC2", DWORD),
        ("dwMaxNumberOfExposuresDESC2", DWORD),
        ("lMinMonitorSignalOffsetDESC2", LONG),
        ("dwMaxMonitorSignalOffsetDESC2", DWORD),
        ("dwMinPeriodicalStepDESC2", DWORD),
        ("dwStartTimeDelayDESC2", DWORD),
        ("dwMinMonitorStepDESC2", DWORD),
        ("dwMinDelayModDESC2", DWORD),
        ("dwMaxDelayModDESC2", DWORD),
        ("dwMinDelayStepModDESC2", DWORD),
        ("dwMinExposureModDESC2", DWORD),
        ("dwMaxExposureModDESC2", DWORD),
        ("dwMinExposureStepModDESC2", DWORD),
        ("dwModulateCapsDESC2", DWORD),
        ("dwReserved", DWORD * 16),
        ("ZZdwDummy", DWORD * 41),
    ]

    def __init__(self):
        super(PCO_Description2, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Description2)


NUM_MAX_SIGNALS = 20
NUM_SIGNALS = 4
NUM_SIGNAL_NAMES = 4


class PCO_Single_Signal_Desc(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("ZZwAlignDummy1", WORD),
        ("strSignalName", (ctypes.c_char * 25) * NUM_SIGNAL_NAMES),
        ("wSignalDefinitions", WORD),
        ("wSignalTypes", WORD),
        ("wSignalPolarity", WORD),
        ("wSignalFilter", WORD),
        ("dwDummy", DWORD * 22),
    ]

    def __init__(self):
        super(PCO_Single_Signal_Desc, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Single_Signal_Desc)


class PCO_Signal_Description(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("wNumOfSignals", WORD),
        ("strSingleSignalDesc", PCO_Single_Signal_Desc * NUM_MAX_SIGNALS),
        ("dwDummy", DWORD * 524),
    ]

    def __init__(self):
        super(PCO_Signal_Description, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Signal_Description)
        for i in range(NUM_MAX_SIGNALS):
            self.strSingleSignalDesc[i].__init__()


PCO_SENSORDUMMY = 7


class PCO_Sensor(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("ZZwAlignDummy1", WORD),
        ("strDescription", PCO_Description),
        ("strDescription2", PCO_Description2),
        ("ZZdwDummy2", DWORD * 256),
        ("wSensorFormat", WORD),
        ("wRoiX0", WORD),
        ("wRoiY0", WORD),
        ("wRoiX1", WORD),
        ("wRoiY1", WORD),
        ("wBinHorz", WORD),
        ("wBinVert", WORD),
        ("ZZwAlignDummy2", WORD),
        ("dwPixelRate", DWORD),
        ("wConvFact", WORD),
        ("wDoubleImage", WORD),
        ("wADCOperation", WORD),
        ("wIR", WORD),
        ("sCoolSet", SHORT),
        ("wOffsetRegulation", WORD),
        ("wNoiseFilterMode", WORD),
        ("wFastReadoutMode", WORD),
        ("wDSNUAdjustMode", WORD),
        ("wCDIMode", WORD),
        ("ZZwDummy", WORD * 36),
        ("strSignalDesc", PCO_Signal_Description),
        ("ZZdwDummy", DWORD * PCO_SENSORDUMMY),
    ]

    def __init__(self):
        super(PCO_Sensor, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Sensor)
        self.strDescription.__init__()
        self.strDescription2.__init__()
        self.strSignalDesc.__init__()

    def __str__(self):
        return (
            str(self.strDescription)
            + """
SensorFormat: {:}
ROI: ({:}, {:}) x ({:}, {:})
PixelRate: {:} Hz""".format(
                self.wSensorFormat,
                self.wRoiX0,
                self.wRoiY0,
                self.wRoiX1,
                self.wRoiY1,
                self.dwPixelRate,
            )
        )


class PCO_Segment(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("wXRes", WORD),
        ("wYRes", WORD),
        ("wBinHorz", WORD),
        ("wBinVert", WORD),
        ("wRoiX0", WORD),
        ("wRoiY0", WORD),
        ("wRoiX1", WORD),
        ("wRoiY1", WORD),
        ("ZZwAlignDummy1", WORD),
        ("dwValidImageCnt", DWORD),
        ("dwMaxImageCnt", DWORD),
        ("wRoiSoftX0", WORD),
        ("wRoiSoftY0", WORD),
        ("wRoiSoftX1", WORD),
        ("wRoiSoftY1", WORD),
        ("wRoiSoftXRes", WORD),
        ("wRoiSoftYRes", WORD),
        ("wRoiSoftDouble", WORD),
        ("ZZwDummy", WORD * 33),
    ]

    def __init__(self):
        super(PCO_Segment, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Segment)


class PCO_Image_ColorSet(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("sSaturation", SHORT),
        ("sVibrance", SHORT),
        ("wColorTemp", WORD),
        ("sTint", SHORT),
        ("wMulNormR", WORD),
        ("wMulNormG", WORD),
        ("wMulNormB", WORD),
        ("sContrast", SHORT),
        ("wGamma", WORD),
        ("wSharpFixed", WORD),
        ("wSharpAdaptive", WORD),
        ("wScaleMin", WORD),
        ("wScaleMax", WORD),
        ("wProcOptions", WORD),
        ("wLutSelection", WORD),
        ("ZZwDummy", WORD * 92),
    ]

    def __init__(self):
        super(PCO_Image_ColorSet, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Image_ColorSet)
        self.wMulNormR = 0x8000
        self.wMulNormG = 0x8000
        self.wMulNormB = 0x8000


class PCO_Image(ctypes.Structure):
    """This class is a binding to the PCO_Image C Structure.
    """

    _fields_ = [
        ("wSize", WORD),
        ("ZZwAlignDummy1", WORD),
        ("strSegment", PCO_Segment * 4),
        ("ZZstrDummySeg", PCO_Segment * 14),
        ("strColorSet", PCO_Image_ColorSet),
        ("wBitAlignment", WORD),
        ("wHotPixelCorrectionMode", WORD),
    ]

    def __init__(self):
        """Constructor method
        """
        super(PCO_Image, self).__init__()
        self.wSize = ctypes.sizeof(PCO_Image)
        for i in range(4):
            self.strSegment[i].__init__()
        for i in range(14):
            self.ZZstrDummySeg[i].__init__()
        self.strColorSet.__init__()


def bcd_byte_to_str(input_):
    """This function converts a one-byte bcd value into two digit string.

    :param input_: bcd value to convert
    :type input_: int
    :return: converted bcd value
    :rtype: str
    """
    if isinstance(input_, Iterable) and len(input_) > 1:
        raise ValueError("Exactly one byte is expected")
    input_a = (int(input_) & 0xF0) >> 4
    input_b = int(input_) & 0x0F
    return "{:d}{:d}".format(input_a, input_b)


def bcd_to_int(input_, endianess="little"):
    """This function converts decimal-coded value (bcd) into int.

    :param input_: input bcd value
    :type input_: byte or bytearray
    :return: integer value
    :rtype: int

    Decimal-encoded value format is given in this table:

    ==============  =====
    Decimal digit   Bits
    ==============  =====
    0               0000
    1               0001
    2               0010
    3               0011
    4               0100
    5               0101
    6               0110
    7               0111
    8               1000
    9               1001
    ==============  =====
    """
    if isinstance(input_, Iterable):
        if endianess == "little":
            return int("".join([bcd_byte_to_str(b) for b in reversed(input_)]))
        else:
            return int("".join([bcd_byte_to_str(b) for b in input_]))
    else:
        return int(bcd_byte_to_str(input_))


class PCO_METADATA(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("wVersion", WORD),
        ("bIMAGE_COUNTER_BCD", BYTE * 4),
        ("bIMAGE_TIME_US_BCD", BYTE * 3),
        ("bIMAGE_TIME_SEC_BCD", BYTE),
        ("bIMAGE_TIME_MIN_BCD", BYTE),
        ("bIMAGE_TIME_HOUR_BCD", BYTE),
        ("bIMAGE_TIME_DAY_BCD", BYTE),
        ("bIMAGE_TIME_MON_BCD", BYTE),
        ("bIMAGE_TIME_YEAR_BCD", BYTE),
        ("bIMAGE_TIME_STATUS", BYTE),
        ("wEXPOSURE_TIME_BASE", WORD),
        ("dwEXPOSURE_TIME", DWORD),
        ("dwFRAMERATE_MILLIHZ", DWORD),
        ("sSENSOR_TEMPERATURE", SHORT),
        ("wIMAGE_SIZE_X", WORD),
        ("wIMAGE_SIZE_Y", WORD),
        ("bBINNING_X", BYTE),
        ("bBINNING_Y", BYTE),
        ("dwSENSOR_READOUT_FREQUENCY", DWORD),
        ("wSENSOR_CONV_FACTOR", WORD),
        ("dwCAMERA_SERIAL_NO", DWORD),
        ("wCAMERA_TYPE", WORD),
        ("bBIT_RESOLUTION", BYTE),
        ("bSYNC_STATUS", BYTE),
        ("wDARK_OFFSET", WORD),
        ("bTRIGGER_MODE", BYTE),
        ("bDOUBLE_IMAGE_MODE", BYTE),
        ("bCAMERA_SYNC_MODE", BYTE),
        ("bIMAGE_TYPE", BYTE),
        ("wCOLOR_PATTERN", WORD),
    ]

    def __init__(self):
        super(PCO_METADATA, self).__init__()
        self.wSize = ctypes.sizeof(PCO_METADATA)

    @property
    def raw_value(self):
        return {key[1:]: getattr(self, key).value for key in self._fields_}

    @property
    def value(self):
        data = self.raw_value
        IMAGE_COUNTER_BCD = data.pop("IMAGE_COUNTER_BCD")
        IMAGE_TIME_US_BCD = data.pop("IMAGE_TIME_BCD")
        IMAGE_TIME_SEC_BCD = data.pop("IMAGE_TIME_SEC_BCD")
        IMAGE_TIME_MIN_BCD = data.pop("IMAGE_TIME_MIN_BCD")
        IMAGE_TIME_HOUR_BCD = data.pop("IMAGE_TIME_HOUR_BCD")
        IMAGE_TIME_DAY_BCD = data.pop("IMAGE_TIME_DAY_BCD")
        IMAGE_TIME_MON_BCD = data.pop("IMAGE_TIME_MON_BCD")
        IMAGE_TIME_YEAR_BCD = data.pop("IMAGE_TIME_YEAR_BCD")

        # IMAGE_COUNTER
        # 0x00000001 to 0x99999999 where first byte is a
        # least significant byte (Little Endian)
        data["IMAGE_COUNTER"] = bcd_to_int(IMAGE_COUNTER_BCD)

        # IMAGE_DATETIME
        data["IMAGE_DATETIME"] = datetime.datetime(
            bcd_to_int(IMAGE_TIME_YEAR_BCD) + 2000,
            bcd_to_int(IMAGE_TIME_MON_BCD),
            bcd_to_int(IMAGE_TIME_DAY_BCD),
            bcd_to_int(IMAGE_TIME_HOUR_BCD),
            bcd_to_int(IMAGE_TIME_MIN_BCD),
            bcd_to_int(IMAGE_TIME_SEC_BCD),
            bcd_to_int(IMAGE_TIME_US_BCD),
        )

        return data


class PCO_Openstruct(ctypes.Structure):
    _fields_ = [
        ("wSize", WORD),
        ("wInterfaceType", WORD),
        ("wCameraNumber", WORD),
        ("wCameraNumAtInterface", WORD),
        ("wOpenFlags", WORD * 10),
        ("dwOpenFlags", DWORD * 5),
        ("wOpenPtr", ctypes.c_void_p * 6),
        ("zzwDummy", WORD * 8),
    ]

    def __init__(self, interface_type=0xFFFF, camera_number=0):
        self.wSize = ctypes.sizeof(PCO_Openstruct)
        inters = {
            "FireWire": 1,
            "Camera Link Matrox": 2,
            "Camera Link Silicon Software mE III": 3,
            "Camera Link National Instruments": 4,
            "GigE": 5,
            "USB 2.0": 6,
            "Camera Link Silicon Software mE IV": 7,
            "USB 3.0": 8,
            "Reserved for WLAN": 9,
            "Camera Link serial port only": 10,
            "Camera Link HS": 11,
            "all": 0xFFFF,
        }
        if interface_type in inters:
            interface_type = inters[interface_type]
        else:
            interface_type = int(interface_type)
        self.wInterfaceType = interface_type
        self.wCameraNumber = camera_number
        self.wCameraNumAtInterface = 0
        for i in range(10):
            self.wOpenFlags[i] = 0
        for i in range(5):
            self.dwOpenFlags[i] = 0
        for i in range(6):
            self.wOpenPtr[i] = 0


# Pixelfly API functions
def PCO_OpenCamera():
    """This function opens a camera device and attach it to a handle, which will be returned by
    the parameter ph. This function scans for the next available camera. If you want to access
    a distinct camera please use PCO_OpenCameraEx.
    Due to historical reasons the wCamNum parameter is a don’t care.
    """

    f = pixelfly_dll.PCO_OpenCamera
    f.argtypes = (ctypes.POINTER(HANDLE), WORD)
    f.restype = ctypes.c_int
    h = HANDLE(0)
    ret_code = f(ctypes.byref(h), 0)  # the argument is ignored.
    PCO_manage_error(ret_code)  # PCO_OpenCamera must be called multiple times
    return h


def PCO_OpenCameraEx(interface_type, camera_number):
    """This function opens a distinct camera, e.g. a camera which is connected to a
    specific interface port.

    :param interface_type: Interface type to look for the camera
    :type interface_type: int
    :param camera_number: ID of the camera
    :type camera_number: int

    The interface_type values are given in this table:

    ===================================   =========================
    Interface                             interface_type
    ===================================   =========================
    FireWire                              1
    Camera Link Matrox                    2
    Camera Link Silicon Software mE III   3
    Camera Link National Instruments      4
    GigE                                  5
    USB 2.0                               6
    Camera Link Silicon Software mE IV    7
    USB 3.0                               8
    Reserved for WLAN                     9
    Camera Link serial port only          10
    Camera Link HS                        11
    all                                   0xFFFF
    ===================================   =========================
    """

    f = pixelfly_dll.PCO_OpenCameraEx
    f.argtypes = (ctypes.POINTER(HANDLE), ctypes.POINTER(PCO_Openstruct))
    f.restype = ctypes.c_int
    h = HANDLE(0)
    strOpenStruct = PCO_Openstruct(interface_type, camera_number)
    ret_code = f(ctypes.byref(h), ctypes.byref(strOpenStruct))
    PCO_manage_error(ret_code)
    return h


def PCO_CloseCamera(handle):
    """This function closes a camera device.

    :param handle: handle of the camera
    :type handle: HANDLE
    """

    f = pixelfly_dll.PCO_CloseCamera
    f.argtypes = (HANDLE,)
    f.restype = ctypes.c_int
    ret_code = f(handle)
    PCO_manage_error(ret_code)


def PCO_GetInfoString(handle):
    """This function reads information about the camera, e.g. firmware versions.

    :param handle: camera handle
    :type handle: HANDLE
    """
    f = pixelfly_dll.PCO_GetInfoString
    f.argtypes = (HANDLE, DWORD, ctypes.c_char_p, WORD)
    f.restype = ctypes.c_int
    p = ctypes.create_string_buffer(256)
    ret_code = f(handle, 0, p, 256)
    PCO_manage_error(ret_code)
    return p.raw.decode("ascii")


def PCO_GetROI(handle: int) -> Tuple[int, int, int, int]:
    """This function returns the current ROI (region of interest) setting in pixels.
    (X0,Y0) is the upper left corner and (X1,Y1) the lower right one.

    :param handle: handle of the camera
    :type handle: HANDLE
    :return: X0, Y0, X1, Y1
    :rtype: int, int, int, int
    """
    f = pixelfly_dll.PCO_GetROI
    f.argtypes = (
        HANDLE,
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
    )
    f.restype = ctypes.c_int
    RoiX0 = WORD()
    RoiY0 = WORD()
    RoiX1 = WORD()
    RoiY1 = WORD()
    ret_code = f(
        handle,
        ctypes.byref(RoiX0),
        ctypes.byref(RoiY0),
        ctypes.byref(RoiX1),
        ctypes.byref(RoiY1),
    )
    PCO_manage_error(ret_code)
    return RoiX0.value, RoiY0.value, RoiX1.value, RoiY1.value


def PCO_SetROI(handle: int, RoiX0: WORD, RoiY0: WORD, RoiX1: WORD, RoiY1: WORD):
    """This function does set a ROI (region of interest) area on the sensor of the camera.

    :param handle: handle of the camera
    :type handle: HANDLE
    :param RoiX0: left border
    :type RoiX0: int
    :param RoiY0: upper border
    :type RoiY0: int
    :param RoiX1: right border
    :type RoiX1: int
    :param RoiY1: lower border
    :type RoiY1: int
    """
    f = pixelfly_dll.PCO_SetROI
    f.argtypes = (HANDLE, WORD, WORD, WORD, WORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, RoiX0, RoiY0, RoiX1, RoiY1)
    PCO_manage_error(ret_code)


def PCO_GetFrameRate(handle):
    """This function returns the current frame rate and exposure
    time settings of the camera.Returned values are only
    valid if last timing command was PCO_SetFrameRate.
    """
    f = pixelfly_dll.PCO_GetFrameRate
    f.argtypes = (
        HANDLE,
        ctypes.POINTER(WORD),
        ctypes.POINTER(DWORD),
        ctypes.POINTER(DWORD),
    )
    f.restype = ctypes.c_int
    FrameRateStatus = WORD()
    FrameRate = DWORD()
    FrameRateExposure = DWORD()
    ret_code = f(
        handle,
        ctypes.byref(FrameRateStatus),
        ctypes.byref(FrameRate),
        ctypes.byref(FrameRateExposure),
    )
    PCO_manage_error(ret_code)
    return FrameRateStatus.value, FrameRate.value, FrameRateExposure.value


def PCO_SetFrameRate(
    handle: int, FrameRateMode: WORD, FrameRate: DWORD, FrameRateExposure: DWORD
):
    """This function sets Frame rate (mHz) and exposure time (ns)
    Frame rate status gives the limiting factors
    if the condition are not met.
    """
    f = pixelfly_dll.PCO_SetFrameRate
    f.argtypes = (
        HANDLE,
        ctypes.POINTER(WORD),
        WORD,
        ctypes.POINTER(DWORD),
        ctypes.POINTER(DWORD),
    )
    f.restype = ctypes.c_int
    FrameRateStatus = WORD()
    dwFrameRate = DWORD(FrameRate)
    dwFrameRateExposure = DWORD(FrameRateExposure)
    ret_code = f(
        handle,
        ctypes.byref(FrameRateStatus),
        FrameRateMode,
        ctypes.byref(dwFrameRate),
        ctypes.byref(dwFrameRateExposure),
    )
    PCO_manage_error(ret_code)
    return FrameRateStatus.value, FrameRate.value, FrameRateExposure.value


def PCO_GetCameraName(handle):
    """This function retrieves the name of the camera.
    """

    f = pixelfly_dll.PCO_GetCameraName
    f.argtypes = (HANDLE, ctypes.c_char_p, WORD)
    f.restype = ctypes.c_int
    cameraName = ctypes.create_string_buffer(41)
    ret_code = f(handle, cameraName, 41)
    PCO_manage_error(ret_code)
    return cameraName.raw.decode("ascii")


def PCO_GetGeneral(handle):
    """This function requests all info contained in the following
    descriptions, especially:

    - camera type, hardware/firmware version, serial number, etc.

    - Request the current camera and power supply temperatures

    """

    f = pixelfly_dll.PCO_GetGeneral
    f.argtypes = (HANDLE, ctypes.POINTER(PCO_General))
    f.restype = ctypes.c_int
    strGeneral = PCO_General()
    ret_code = f(handle, ctypes.byref(strGeneral))
    PCO_manage_error(ret_code)
    return strGeneral


def PCO_GetSensorStruct(handle):
    """Get the complete set of the sensor functions settings
    """

    f = pixelfly_dll.PCO_GetSensorStruct
    f.argtypes = (HANDLE, ctypes.POINTER(PCO_Sensor))
    f.restype = ctypes.c_int
    strSensor = PCO_Sensor()
    ret_code = f(handle, ctypes.byref(strSensor))
    PCO_manage_error(ret_code)
    return strSensor


def PCO_GetCameraDescription(handle):
    """Sensor and camera specific description is queried. In the returned
    PCO_Description structure margins for all sensor related settings
    and bitfields for available options of the camera are given.
    """

    f = pixelfly_dll.PCO_GetCameraDescription
    f.argtypes = (HANDLE, ctypes.POINTER(PCO_Description))
    f.restype = ctypes.c_int
    desc = PCO_Description()
    ret_code = f(handle, ctypes.byref(desc))
    PCO_manage_error(ret_code)
    return desc


def PCO_GetCameraHealthStatus(handle):
    """This function retrieves information about the current camera status.
    """

    # Call PCO_GetCameraHealthStatus function
    f = pixelfly_dll.PCO_GetCameraHealthStatus
    f.argtypes = (
        HANDLE,
        ctypes.POINTER(DWORD),
        ctypes.POINTER(DWORD),
        ctypes.POINTER(DWORD),
    )
    f.restype = ctypes.c_int
    dwWarn = DWORD()
    dwErr = DWORD()
    dwStatus = DWORD()
    ret_code = f(
        handle, ctypes.byref(dwWarn), ctypes.byref(dwErr), ctypes.byref(dwStatus)
    )
    PCO_manage_error(ret_code)

    return dwWarn.value, dwErr.value, dwStatus.value


def PCO_GetRecordingStruct(handle):
    """Get the complete set of the recording function
    settings. Please fill in all wSize parameters,
    even in embedded structures.
    """

    strRecording = PCO_Recording()
    f = pixelfly_dll.PCO_GetRecordingStruct
    f.argtypes = (HANDLE, ctypes.POINTER(PCO_Recording))
    f.restype = ctypes.c_int
    ret_code = f(handle, ctypes.byref(strRecording))
    PCO_manage_error(ret_code)
    return strRecording


def PCO_GetSizes(handle: int) -> Tuple[int, int, int, int]:
    """This function returns the current armed image size of the camera.
    """

    f = pixelfly_dll.PCO_GetSizes
    f.argtypes = (
        HANDLE,
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
    )
    f.restype = ctypes.c_int
    XResAct = WORD()
    YResAct = WORD()
    XResMax = WORD()
    YResMax = WORD()
    ret_code = f(
        handle,
        ctypes.byref(XResAct),
        ctypes.byref(YResAct),
        ctypes.byref(XResMax),
        ctypes.byref(YResMax),
    )
    PCO_manage_error(ret_code)
    return XResAct.value, YResAct.value, XResMax.value, YResMax.value


def PCO_AllocateBuffer(
    handle: int, bufNr: int, size: int, bufPtr: ctypes.POINTER(WORD), hEvent: HANDLE = 0
) -> Tuple[int, ctypes.POINTER(WORD), int]:
    """This function sets up a buffer context to receive the transferred
    images. A buffer index is returned, which must be used for the
    image transfer functions. There is a maximum of 16 buffers per camera.

    Attention: This function cannot be used, if the connection to the
    camera is established through the serial connection of a Camera
    Link grabber. In this case, the SDK of the grabber must be used
    to do any buffer management.

    .. todo::
        est-ce le cas avec le grabber Solios ?
    """

    f = pixelfly_dll.PCO_AllocateBuffer
    f.argtypes = (
        HANDLE,  # handle
        ctypes.POINTER(SHORT),  # sBufNr
        DWORD,  # dwSize
        ctypes.POINTER(ctypes.POINTER(WORD)),
        ctypes.POINTER(HANDLE),
    )
    f.restype = ctypes.c_int
    sBufNr = SHORT(bufNr)
    hEvent = HANDLE(hEvent)
    ret_code = f(
        handle, ctypes.byref(sBufNr), size, ctypes.byref(bufPtr), ctypes.byref(hEvent)
    )
    PCO_manage_error(ret_code)
    return sBufNr.value, hEvent.value


def PCO_FreeBuffer(handle, bufNr):
    """This function frees a previously allocated buffer context with
    a given index. If internal memory was allocated for this buffer
    context, it will be freed. If an internal event handle was
    created, it will be closed.
    """

    f = pixelfly_dll.PCO_FreeBuffer
    f.argtypes = (HANDLE, SHORT)
    f.restype = ctypes.c_int
    ret_code = f(handle, bufNr)
    PCO_manage_error(ret_code)


def PCO_GetBufferStatus(handle, sBufNr):
    """This function queries the status of the buffer context with the given index:

    - StatusDll describes the state of the buffer context:
        ==========  ==================================================
        StatusDll   description
        ==========  ==================================================
        0x80000000  Buffer is allocated
        0x40000000  Buffer event created inside the SDK DLL
        0x20000000  Buffer is allocated externally
        0x00008000  Buffer event is set
        ==========  ==================================================

    - StatusDrv describes the state of the last image transfer into this buffer.
        * PCO_NOERROR: Image transfer succeeded

        * others: See error codes

    """

    f = pixelfly_dll.PCO_GetBufferStatus
    f.argtypes = (HANDLE, SHORT, ctypes.POINTER(DWORD), ctypes.POINTER(DWORD))
    f.restype = ctypes.c_int
    statusDLL = DWORD()
    statusDrv = DWORD()
    ret_code = f(handle, sBufNr, ctypes.byref(statusDLL), ctypes.byref(statusDrv))
    PCO_manage_error(ret_code)
    return statusDLL.value, statusDrv.value


def PCO_ArmCamera(handle):
    """Arms, i.e. prepares the camera for a consecutive
    set recording status = [run] command. All
    configurations and settings made up to this moment
    are accepted and the internal settings of the
    camera is prepared. Thus the camera is able to
    start immediately when the set recording status
    = [run] command is performed.
    """

    f = pixelfly_dll.PCO_ArmCamera
    f.argtypes = (HANDLE,)
    f.restype = ctypes.c_int
    ret_code = f(handle)
    PCO_manage_error(ret_code)


def PCO_GetRecordingState(handle):
    """Returns the current Recording state of the camera:
    - 0x0000: camera is stopped, recording state [stop]

    - 0x0001: camera is running, recording state [run]

    """

    f = pixelfly_dll.PCO_GetRecordingState
    f.argtypes = (HANDLE, ctypes.POINTER(WORD))
    f.restype = ctypes.c_int
    state = WORD()
    ret_code = f(handle, ctypes.byref(state))
    PCO_manage_error(ret_code)
    return state.value


def PCO_SetRecordingState(handle, state):
    """Sets the current recording status and waits till
    the status is valid. If the state can't be set the
    function will return an error.

    .. note::
        - it is necessary to arm camera before every set recording status command in order to ensure that all settings are accepted correctly.

        - During the recording session, it is possible to change the timing by calling :func:`PCO_SetDelayExposureTime`.
    """

    if state not in (0x0000, 0x0001):
        raise ValueError("Wrong state value")
    f = pixelfly_dll.PCO_SetRecordingState
    f.argtypes = (HANDLE, WORD)
    f.restype = ctypes.c_int
    wRecState = WORD(1 if state else 0)
    ret_code = f(handle, wRecState)
    PCO_manage_error(ret_code)


def PCO_GetBitAlignment(handle):
    """This function returns the current bit alignment of the
    transferred image data. The data can be either
    MSB (Big Endian) or LSB (Little Endian) aligned.
    Returns:
    - 0x0000 [MSB]

    - 0x0001 [LSB]

    """

    f = pixelfly_dll.PCO_GetBitAlignment
    f.argtypes = (HANDLE, ctypes.POINTER(WORD))
    f.restype = ctypes.c_int
    bitAlignment = WORD()
    ret_code = f(handle, ctypes.byref(bitAlignment))
    PCO_manage_error(ret_code)
    return bitAlignment.value


def PCO_SetBitAlignment(handle, littleEndian):
    """This functions sets the bit alignment of the transferred
    image data. littleEndian can be 0 or 1.
    """

    f = pixelfly_dll.PCO_SetBitAlignment
    f.argtypes = (HANDLE, WORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, littleEndian)
    PCO_manage_error(ret_code)


def PCO_GetImageStruct(handle):
    """Information about previously recorded images is queried from
    the camera and the variables of the PCO_Image structure are
    filled with this information.
    """

    f = pixelfly_dll.PCO_GetImageStruct
    f.argtypes = (HANDLE, ctypes.POINTER(PCO_Image))
    f.restype = ctypes.c_int
    strImage = PCO_Image()
    ret_code = f(handle, ctypes.byref(strImage))
    PCO_manage_error(ret_code)
    return strImage


def PCO_GetMetaData(handle, bufNr):
    """Cameras: pco.dimax and pco.edge

    Query additionnal image information, which the camera has attached to
    the transferred image, if Meta Data mode is enabled.
    """

    f = pixelfly_dll.PCO_GetMetaData
    f.argtypes = (HANDLE, SHORT, ctypes.POINTER(PCO_METADATA), DWORD, DWORD)
    f.restype = ctypes.c_int
    MetaData = PCO_METADATA()
    ret_code = f(handle, bufNr, ctypes.byref(MetaData), 0, 0)
    PCO_manage_error(ret_code)
    return MetaData.value


def PCO_SetMetaDataMode(handle, MetaDataMode):
    """Cameras: pco.dimax and pco.edge

    Sets the mode for Meta Data and returns information about size and version
    of the Meta Data block.
    When Meta Data mode is set to [on], a Meta Data block with additional information
    is added at the end of each image. The internal buffers allocated with PCO_AllocateBuffer
    are adapted automatically.
    """

    f = pixelfly_dll.PCO_SetMetaDataMode
    f.argtypes = (HANDLE, WORD, ctypes.POINTER(WORD), ctypes.POINTER(WORD))
    f.restype = ctypes.c_int
    MetaDataSize = WORD()
    MetaDataVersion = WORD()
    ret_code = f(
        handle, MetaDataMode, ctypes.byref(MetaDataSize), ctypes.byref(MetaDataVersion)
    )
    PCO_manage_error(ret_code)
    return MetaDataSize.value, MetaDataVersion.value


def PCO_GetMetaDataMode(handle):
    """Returns the current Meta Data mode of the camera and information about size
    and version of the Meta Data block.
    """

    f = pixelfly_dll.PCO_GetMetaDataMode
    f.argtypes = (
        HANDLE,
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
    )
    f.restype = ctypes.c_int
    MetaDataMode = WORD()
    MetaDataSize = WORD()
    MetaDataVersion = WORD()
    ret_code = f(
        handle,
        ctypes.byref(MetaDataMode),
        ctypes.byref(MetaDataSize),
        ctypes.byref(MetaDataVersion),
    )
    PCO_manage_error(ret_code)
    return MetaDataMode.value, MetaDataSize.value, MetaDataVersion.value


def PCO_SetTimestampMode(handle, mode):
    """Sets the timestamp mode of the camera:

    ======= =================== =========================================================
    mode    short description   long description
    ======= =================== =========================================================
    0x0000  [off]
    0x0001  [binary]            BCD coded timestamp in the first 14 pixels
    0x0002  [binary+ASCII]      BCD coded timestamp in the first 14 pixels + ASCII text
    0x0003  [ASCII]             ASCII text only (see camera descriptor for availability)
    ======= =================== =========================================================
    """

    if mode not in (0x0000, 0x0001, 0x0002, 0x0003):
        raise ValueError("Bad mode value")
    f = pixelfly_dll.PCO_SetTimestampMode
    f.argtypes = (HANDLE, WORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, mode)
    PCO_manage_error(ret_code)


def PCO_AddBufferEx(
    handle, dw1stImage, dwLastImage, sBufNr, wXRes, wYRes, wBitPerPixel
):
    """This function sets up a request for a single transfer from the camera and
    returns immediately.
    """

    f = pixelfly_dll.PCO_AddBufferEx
    f.argtypes = (HANDLE, DWORD, DWORD, SHORT, WORD, WORD, WORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, dw1stImage, dwLastImage, sBufNr, wXRes, wYRes, wBitPerPixel)
    PCO_manage_error(ret_code)


def PCO_CancelImages(handle):
    """This function does remove all remaining buffers from the internal
    queue, reset the internal queue and also reset the transfer state
    machine in the camera. It is mandatory to call PCO_CancelImages
    after all image transfers are done. This function can be called
    before or after setting PCO_SetRecordingState to [stop].
    """

    f = pixelfly_dll.PCO_CancelImages
    f.argtypes = (HANDLE,)
    f.restype = ctypes.c_int
    ret_code = f(handle)
    PCO_manage_error(ret_code)


IMAGEPARAMETERS_READ_WHILE_RECORDING = 0x00000001
IMAGEPARAMETERS_READ_FROM_SEGMENTS = 0x00000002


def PCO_SetImageParameters(handle, XRes, YRes, flags):
    """This function sets the image parameters for internal allocated resources.
    This function must be called before an image transfer is started.
    If next image will be transfered from a recording camera, flag
    IMAGEPARAMETERS_READ_WHILE_RECORDING must be set. If next action is to
    readout images from the camera internal memory, flag
    IMAGEPARAMETERS_READ_FROM_SEGMENTS must be set.
    """

    if flags not in (
        IMAGEPARAMETERS_READ_WHILE_RECORDING,
        IMAGEPARAMETERS_READ_FROM_SEGMENTS,
    ):
        raise ValueError("Wrong flag value")

    f = pixelfly_dll.PCO_SetImageParameters
    f.argtypes = (HANDLE, WORD, WORD, DWORD, ctypes.c_void_p, ctypes.c_int)
    f.restype = ctypes.c_int
    ret_code = f(handle, XRes, YRes, flags, ctypes.c_void_p(), 0)
    PCO_manage_error(ret_code)


def PCO_GetImageEx(
    handle, segment, firstImage, lastImage, bufNr, xRes, yRes, bitsPerPixel
):
    """This function can be used to get a single image from the camera.
    The function does not return until the image is transferred to the
    buffer or an error occured. The timeout value for the transfer
    can be set with function PCO_SetTimeouts, the default value is
    6 seconds. On return the image stored in the memory area of the
    buffer, which is addressed through parameter sBufNr.
    """

    f = pixelfly_dll.PCO_GetImageEx
    f.argtypes = (
        HANDLE,  # handle
        WORD,  # wSegment
        DWORD,  # dw1stImage
        DWORD,  # dwLastImage
        SHORT,  # sBufNr
        WORD,  # wXRes
        WORD,  # wYRes
        WORD,  # wBitPerPixel
    )
    f.restype = ctypes.c_int
    ret_code = f(
        handle, segment, firstImage, lastImage, bufNr, xRes, yRes, bitsPerPixel
    )
    PCO_manage_error(ret_code)


PCO_Timebases = {0x0000: 1e-9, 0x0001: 1e-6, 0x0002: 1e-3}


def PCO_SetDelayExposureTime(
    handle, dwDelay, dwExposure, wTimeBaseDelay, wTimeBaseExposure
):
    """This function sets the delay and exposure time and the
    associated time base values.
    Restrictions for the parameter values are defined in the
    PCO_Description structure:
    - dwMinDelayDESC

    - dwMaxDelayDESC

    - dwMinDelayStepDESC

    - dwMinExposDESC

    - dwMaxExposDESC

    - dwMinExposStepDESC

    Possible values for wTimeBaseDelay and wTimeBaseExposure:

    ======= =====
    Value   Unit
    ======= =====
    0x0000  ns
    0x0001  µs
    0x0002  ms
    ======= =====

    """

    f = pixelfly_dll.PCO_SetDelayExposureTime
    f.argtypes = (HANDLE, DWORD, DWORD, WORD, WORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, dwDelay, dwExposure, wTimeBaseDelay, wTimeBaseExposure)
    PCO_manage_error(ret_code)


def PCO_GetDelayExposureTime(handle):
    """Returns the current setting of delay and exposure time
    """

    f = pixelfly_dll.PCO_GetDelayExposureTime
    f.argtypes = (
        HANDLE,
        ctypes.POINTER(DWORD),
        ctypes.POINTER(DWORD),
        ctypes.POINTER(WORD),
        ctypes.POINTER(WORD),
    )
    f.restype = ctypes.c_int
    delay = DWORD()
    exposure = DWORD()
    timebase_delay = WORD()
    timebase_exposure = WORD()
    ret_code = f(
        handle,
        ctypes.byref(delay),
        ctypes.byref(exposure),
        ctypes.byref(timebase_delay),
        ctypes.byref(timebase_exposure),
    )
    PCO_manage_error(ret_code)
    return delay.value, exposure.value, timebase_delay.value, timebase_exposure.value


def PCO_GetTriggerMode(handle):
    """Returns the current trigger mode setting of the
    camera
    """

    f = pixelfly_dll.PCO_GetTriggerMode
    f.argtypes = (HANDLE, ctypes.POINTER(WORD))
    f.restype = ctypes.c_int
    triggerMode = WORD()
    ret_code = f(handle, ctypes.byref(triggerMode))
    PCO_manage_error(ret_code)
    return triggerMode.value


PCO_TriggerModeDescription = {
    0x0000: "auto sequence",
    0x0001: "software trigger",
    0x0002: "external exposure start & software trigger",
    0x0003: "external exposure control",
    0x0004: "external synchronized",
    0x0005: "fast external exposure control",
    0x0006: "external CDS control",
    0x0007: "slow external exposure control",
    0x0102: "external synchronized HDSDI",
}


def PCO_SetTriggerMode(handle, mode):
    """Sets the trigger mode of the camera.
    """

    f = pixelfly_dll.PCO_SetTriggerMode
    f.argtypes = (HANDLE, WORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, mode)
    PCO_manage_error(ret_code)


def PCO_SetADCOperation(handle, operation):
    """Sets the ADC (analog-digital-converter) operating mode.
    If sensor data is read out using single ADC operation,
    linearity of image data is enhanced. Using dual ADC,
    operation readout is faster and allows higher frame rates.
    If dual ADC operating mode is set, horizontal ROI must be
    adapted to symmetrical values.

    Possible values:

    ======= =============
    Value   Mode
    ======= =============
    0x0001  [single ADC]
    0x0002  [dual ADC]
    ======= =============

    """

    f = pixelfly_dll.PCO_SetADCOperation
    f.argtypes = (HANDLE, WORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, operation)
    PCO_manage_error(ret_code)


def PCO_GetADCOperation(handle):
    """Returns the ADC operation mode (single / dual)
    """

    f = pixelfly_dll.PCO_GetADCOperation
    f.argtypes = (HANDLE, ctypes.POINTER(WORD))
    f.restype = ctypes.c_int
    operation = WORD()
    ret_code = f(handle, ctypes.byref(operation))
    PCO_manage_error(ret_code)
    return operation.value


def PCO_SetPixelRate(handle, rate):
    """This functions sets the pixel rate for the sensor readout.
    """

    f = pixelfly_dll.PCO_SetPixelRate
    f.argtypes = (HANDLE, DWORD)
    f.restype = ctypes.c_int
    ret_code = f(handle, rate)
    PCO_manage_error(ret_code)


def PCO_GetPixelRate(handle):
    """Returns the current pixel rate of the camera in Hz.
    The pixel rate determines the sensor readout speed.
    """

    f = pixelfly_dll.PCO_GetPixelRate
    f.argtypes = (HANDLE, ctypes.POINTER(DWORD))
    f.restype = ctypes.c_int
    rate = DWORD()
    ret_code = f(handle, ctypes.byref(rate))
    PCO_manage_error(ret_code)
    return rate.value


def PCO_GetNoiseFilterMode(handle):
    """This function returns the current operating mode of the image correction
    in the camera.

    The noise filter mode:

    ======= =============================
    Value   Mode
    ======= =============================
    0x0000  [off]
    0x0001  [on]
    0x0101  [on + hot pixel correction]
    ======= =============================
    """

    f = pixelfly_dll.PCO_GetNoiseFilterMode
    f.argtypes = (HANDLE, ctypes.POINTER(WORD))
    f.restype = ctypes.c_int
    noiseFilterMode = WORD()
    ret_code = f(handle, ctypes.byref(noiseFilterMode))
    PCO_manage_error(ret_code)
    return noiseFilterMode.value


def PCO_SetNoiseFilterMode(handle, mode):
    """This function does set the image correction operating mode of the camera. Image correction can either be switched to totally off,
    noise filter only mode or noise filter plus hot pixel correction mode.
    The command will be rejected, if Recording State is [run], see PCO_GetRecordingState.

    The noise filter mode:

    ======= =============================
    Value   Mode
    ======= =============================
    0x0000  [off]
    0x0001  [on]
    0x0101  [on + hot pixel correction]
    ======= =============================

    """
    if mode not in (0x0000, 0x0001, 0x0101):
        raise ValueError("Wrong mode value")
    f = pixelfly_dll.PCO_SetNoiseFilterMode
    f.argtypes = (HANDLE, WORD)
    f.restype = ctypes.c_int
    ret_type = f(handle, mode)
    PCO_manage_error(ret_type)


if __name__ == "__main__":
    try:
        h = PCO_OpenCamera()
        print("h =", h)
        info = PCO_GetInfoString(h)
        print("info =", info)
        general = PCO_GetGeneral(h)
        print("general =", general)
        sensor = PCO_GetSensorStruct(h)
        print("sensor =", sensor)
    except PCO_Error as pe:
        print('Error: "', pe.args[0], '"')
    finally:
        PCO_CloseCamera(h)
        print("Camera closed")



"""Asynchronous Session Module (:mod:`pymanip.asyncsession`)
============================================================

This module defines two classes for live acquisition, :class:`~pymanip.asyncsession.AsyncSession`
and :class:`~pymanip.asyncsession.RemoteObserver`. The former is used to
manage an experimental session, the latter to access its live data from a
remote computer. There is also one class for read-only access to previous session,
:class:`~pymanip.asyncsession.SavedAsyncSession`.

.. autoclass:: AsyncSession
   :members:
   :private-members:

.. autoclass:: SavedAsyncSession
   :members:
   :private-members:

.. autoclass:: RemoteObserver
   :members:
   :private-members:

.. _FluidLab: https://foss.heptapod.net/fluiddyn/fluidlab

"""

import signal
import time
import sys
import os.path
import pickle
import warnings
import inspect
from pprint import pprint
from functools import lru_cache

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property

import sqlite3
from datetime import datetime
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.cbook import mplDeprecation as MatplotlibDeprecationWarning

import asyncio
from aiohttp import web
import aiohttp_jinja2
import jinja2
import tempfile
import smtplib
from email.message import EmailMessage
import requests
import json

try:
    import PyQt5.QtCore
except (ModuleNotFoundError, FileNotFoundError):
    pass

from fluiddyn.util.terminal_colors import cprint
from pymanip.mytime import dateformat

__all__ = ["AsyncSession"]


class AsyncSession:
    """This class represents an asynchronous experiment session. It is the main tool that we
    use to set up monitoring of experimental systems. It will manage the storage for the data,
    as well as several asynchronous functions for use during the monitoring of the experimental
    system such as live plot of monitored data, regular control email, and remote HTTP access to
    the live data by human (connection from a web browser), or by a machine using the
    :class:`pymanip.asyncsession.RemoteObserver` class. It has methods to access the data for
    processing during the experiment, or post-processing after the experiment is finished.

    :param session_name: the name of the session, defaults to None. It will be used as the filename of the sqlite3 database file stored on disk. If None, then no file is created, and data will be temporarily stored in memory, but will be lost when the object is released.
    :type session_name: str, optional
    :param verbose: sets the session verbosity, defaults to True
    :type verbose: bool, optional
    :param delay_save: if True, the data is stored in memory during the duration of the session, and is saved to disk only at the end of the session. It is not recommanded, but useful in cases where fast operation requires to avoid disk access during the session.
    :type delay_save: bool, optional
    """

    database_version = 3

    def __init__(
        self,
        session_name=None,
        verbose=True,
        delay_save=False,
        exist_ok=True,
        readonly=False,
    ):
        """Constructor method"""

        if session_name is not None:
            session_name = str(session_name)  # in case it is a Path object
            if session_name.endswith(".db"):
                session_name = session_name[:-3]
        elif delay_save:
            raise ValueError("Cannot delay_save if session_name is not specified")

        self.session_name = session_name
        self.verbose = verbose
        self.exist_ok = exist_ok
        self.readonly = readonly
        self.delay_save = delay_save

        self.custom_figures = None
        self.figure_list = []
        self.template_dir = os.path.join(os.path.dirname(__file__), "web")
        self.static_dir = os.path.join(os.path.dirname(__file__), "web_static")
        self.jinja2_loader = jinja2.FileSystemLoader(self.template_dir)
        self.conn = None

    def save_database(self):
        """This method is useful only if delay_save = True. Then, the database is kept in-memory for
        the duration of the session. This method saves the database on the disk.
        A new database file will be created with the content of the current
        in-memory database.

        This method is automatically called at the exit of the context manager.
        """
        if self.conn is None:
            raise RuntimeError("AsyncSession is not opened!")

        if self.delay_save:
            try:
                os.remove(str(self.session_name) + ".db")
            except FileNotFoundError:
                pass
            disk_db = sqlite3.connect(str(self.session_name) + ".db")
            try:
                with disk_db as c:
                    for line in self.conn.iterdump():
                        c.execute(line)
            finally:
                disk_db.close()

    def open(self):
        """Opens database for reading or writting"""

        if self.conn is not None:
            raise RuntimeError("Already opened!")

        if not self.exist_ok and os.path.exists(self.session_name + ".db"):
            raise RuntimeError("File exists !")

        if self.session_name is None or self.delay_save:
            # For no name session, or in case of delay_save=True, then
            # the connection is in-memory
            self.conn = sqlite3.connect(":memory:")
        else:
            # Otherwise, the connection is on the disk for immediate read/write
            if self.readonly:
                uri = "file:{path:}?mode=ro".format(path=self.session_name + ".db")
                self.conn = sqlite3.connect(uri, uri=True)
            else:
                self.conn = sqlite3.connect(self.session_name + ".db")

        if self.delay_save and os.path.exists(self.session_name + ".db"):
            # Load existing database into in-memory database
            uri = "file:{path:}?mode=ro".format(path=self.session_name + ".db")
            disk_db = sqlite3.connect(uri, uri=True)
            try:
                with self.conn as c:
                    for line in disk_db.iterdump():
                        c.execute(line)
            finally:
                disk_db.close()

        with self.conn as c:
            tables = list(c.execute("SELECT name FROM sqlite_master;"))
            if not tables:
                c.execute(
                    """
                    CREATE TABLE log_names (
                    name TEXT);
                    """
                )
                c.execute(
                    """
                    CREATE TABLE log (
                    timestamp INT,
                    name TEXT,
                    value REAL);
                    """
                )
                c.execute(
                    """
                    CREATE TABLE dataset_names (
                    name TEXT);
                    """
                )
                c.execute(
                    """
                    CREATE TABLE dataset (
                    timestamp INT,
                    name TEXT,
                    data BLOB);
                    """
                )
                c.execute(
                    """
                    CREATE TABLE parameters (
                        name TEXT,
                        value REAL);
                    """
                )
                c.execute(
                    """
                    INSERT INTO parameters
                    (name, value)
                    VALUES (?,?);
                    """,
                    ("_database_version", AsyncSession.database_version),
                )
                c.execute(
                    """
                    INSERT INTO parameters
                    (name, value)
                    VALUES (?,?);
                    """,
                    ("_session_creation_timestamp", datetime.now().timestamp()),
                )
            elif self.verbose:
                self.print_welcome()

    def close(self):
        if self.conn:
            self.save_database()
            self.conn.close()
        self.conn = None

    def __enter__(self):
        """Context manager enter method"""
        if not self.conn:
            self.open()
        return self

    def __exit__(self, type_, value, cb):
        """Context manager exit method"""
        self.close()

    def get_version(self):
        """Returns current version of the database layout."""
        version = self.parameter("_database_version")
        if version is None:
            version = 1
        return version

    @cached_property
    def t0(self):
        """Session creation timestamp"""
        t0 = self.parameter("_session_creation_timestamp")
        if t0 is not None:
            return t0
        logged_data = self.logged_first_values()
        if logged_data:
            t0 = min([v[0] for k, v in logged_data.items()])
            self.save_parameter(_session_creation_timestamp=t0)
            return t0
        return 0

    @property
    def initial_timestamp(self):
        """Session creation timestamp, identical to :attr:`pymanip.asyncsession.AsyncSession.t0`"""
        return self.t0

    @property
    def last_timestamp(self):
        """Timestamp of the last recorded value"""
        ts = list()
        last_values = self.logged_last_values()
        if last_values:
            ts.append(max([t_v[0] for name, t_v in last_values.items()]))
        for ds_name in self.dataset_names():
            ts.append(max(self.dataset_times(ds_name)))
        if ts:
            return max(ts)
        return None

    def print_welcome(self):
        """Prints informative start date/end date message. If verbose is True, this method
        is called by the constructor.
        """
        start_string = time.strftime(dateformat, time.localtime(self.initial_timestamp))
        cprint.blue("*** Start date: " + start_string)
        last = self.last_timestamp
        if last:
            end_string = time.strftime(dateformat, time.localtime(last))
            cprint.blue("***   End date: " + end_string)

    def print_description(self):
        """Prints the list of parameters, logged variables and datasets."""
        version = self.get_version()
        print(
            self.session_name,
            "is an asynchroneous session (version {:}).".format(version),
        )
        print()
        last_values = self.logged_last_values()
        params = {
            key: val
            for key, val in self.parameters().items()
            if not key.startswith("_")
        }
        if params:
            print("Parameters")
            print("==========")
            for key, val in self.parameters().items():
                print(key, ":", val)
            print()

        if last_values:
            print("Logged variables")
            print("================")
            for name, t_v in last_values.items():
                print(name, "(", t_v[1], ")")
            print()

        ds_names = self.dataset_names()
        if ds_names:
            print("Datasets")
            print("========")
            for ds in ds_names:
                print(ds)
            print()

    def add_entry(self, *args, **kwargs):
        """This methods adds scalar values into the database. Each entry value
        will hold a timestamp corresponding to the time at which this method has been called.
        Variables are passed in dictionnaries or as keyword-arguments. If several variables
        are passed, then they all hold the same timestamps.

        For parameters which consists in only one scalar value, and for which timestamps are
        not necessary, use :meth:`pymanip.asyncsession.AsyncSession.save_parameter` instead.

        :param \\*args: dictionnaries with name-value to be added in the database
        :type \\*args: dict, optional
        :param \\**kwargs: name-value to be added in the database
        :type \\**kwargs: float, optional
        """
        ts = datetime.now().timestamp()
        data = dict()
        for a in args:
            data.update(a)
        data.update(kwargs)
        with self.conn as c:
            cursor = c.cursor()
            cursor.execute("SELECT name FROM log_names;")
            names = set([d[0] for d in cursor.fetchall()])
            for key, val in data.items():
                if key not in names:
                    c.execute("INSERT INTO log_names VALUES (?);", (key,))
                    names.add(key)
                c.execute("INSERT INTO log VALUES (?,?,?);", (ts, key, val))

    def add_dataset(self, *args, **kwargs):
        """This method adds arrays, or other pickable objects, as “datasets” into the
        database. They will hold a timestamp corresponding to the time at which the method
        has been called.

        :param \\*args: dictionnaries with name-value to be added in the database
        :type \\*args: dict, optional
        :param \\**kwargs: name-value to be added in the database
        :type \\**kwargs: object, optional
        """
        ts = datetime.now().timestamp()
        data = dict()
        for a in args:
            data.update(a)
        data.update(kwargs)
        with self.conn as c:
            cursor = c.cursor()
            cursor.execute("SELECT name FROM dataset_names;")
            names = set([d[0] for d in cursor.fetchall()])
            for key, val in data.items():
                if key not in names:
                    c.execute("INSERT INTO dataset_names VALUES (?);", (key,))
                    names.add(key)
                c.execute(
                    "INSERT INTO dataset VALUES (?,?,?);",
                    (ts, key, pickle.dumps(val, protocol=4)),
                )

    def logged_variables(self):
        """This method returns a set of the names of the scalar variables currently stored
        in the session database.

        :return: names of scalar variables
        :rtype: set
        """
        with self.conn as conn:
            c = conn.cursor()
            c.execute("SELECT name FROM log_names;")
            data = c.fetchall()
        names = set([d[0] for d in data])
        return names

    def logged_data(self):
        """This method returns a name-value dictionnary containing all scalar variables
        currently stored in the session database.

        :return: all scalar variable values
        :rtype: dict
        """
        names = self.logged_variables()
        result = dict()
        for name in names:
            result[name] = self.__getitem__(name)
        return result

    def logged_variable(self, varname):
        """This method retrieve the timestamps and values of a specified scalar variable.
        It is possible to use the sesn[varname] syntax as a shortcut.

        :param varname: name of the scalar variable to retrieve
        :type varname: str
        :return: timestamps and values
        :rtype: tuple (timestamps, values) of numpy arrays.

        :Exemple:

        >>> ts, val = sesn.logged_variable('T_Pt_bas')

        """
        with self.conn as conn:
            c = conn.cursor()
            c.execute(
                """
                      SELECT timestamp, value FROM log
                      WHERE name='{:}';
                      """.format(
                    varname
                )
            )
            data = c.fetchall()
        t = np.array([d[0] for d in data])
        v = np.array([d[1] for d in data])
        return t, v

    def logged_first_values(self):
        """This method returns a dictionnary holding the first logged value of all scalar
        variables stored in the session database.

        :return: first values
        :rtype: dict
        """
        with self.conn as conn:
            c = conn.cursor()
            c.execute("SELECT name FROM log_names;")
            names = set([d[0] for d in c.fetchall()])
            result = dict()
            for name in names:
                c.execute(
                    """SELECT timestamp, value FROM log
                             WHERE name='{:}'
                             ORDER BY timestamp ASC
                             LIMIT 1;
                          """.format(
                        name
                    )
                )
                result[name] = c.fetchone()
        return result

    def logged_last_values(self):
        """This method returns a dictionnary holding the last logged value of all scalar
        variables stored in the session database.

        :return: last logged values
        :rtype: dict
        """
        with self.conn as conn:
            c = conn.cursor()
            c.execute("SELECT name FROM log_names;")
            names = set([d[0] for d in c.fetchall()])
            result = dict()
            for name in names:
                c.execute(
                    """SELECT timestamp, value FROM log
                             WHERE name='{:}'
                             ORDER BY timestamp DESC
                             LIMIT 1;
                          """.format(
                        name
                    )
                )
                result[name] = c.fetchone()
        return result

    def logged_data_fromtimestamp(self, name, timestamp):
        """This method returns the timestamps and values of a given scalar variable, recorded
        after the specified timestamp.

        :param name: name of the scalar variable to be retrieved.
        :type name: str
        :param timestamp: timestamp from which to recover values
        :type timestamp: float
        :return: the timestamps, and values of the specified variable
        :rtype: tuple of two numpy arrays
        """
        with self.conn as conn:
            c = conn.cursor()
            c.execute(
                """SELECT timestamp, value FROM log
                         WHERE name='{:}' AND timestamp > {:}
                         ORDER BY timestamp ASC;
                      """.format(
                    name, timestamp
                )
            )
            data = c.fetchall()
        t = np.array([d[0] for d in data if d[1] is not None])
        v = np.array([d[1] for d in data if d[1] is not None])
        return t, v

    def dataset_names(self):
        """This method returns the names of the datasets currently stored in the session
        database.

        :return: names of datasets
        :rtype: set
        """
        with self.conn as conn:
            c = conn.cursor()
            try:
                c.execute("SELECT name from dataset_names;")
                data = c.fetchall()
            except sqlite3.OperationalError:
                return set()
        return set([d[0] for d in data])

    def datasets(self, name):
        """This method returns a generator which will yield all timestamps and datasets
        recorded under the specified name.
        The rationale for returning a generator instead of a list, is that each individual
        dataset may be large.

        :param name: name of the dataset to retrieve
        :type name: str

        :Exemple:

        - To plot all the recorded datasets named 'toto'

        >>> for timestamp, data in sesn.datasets('toto'):
        >>>    plt.plot(data, label=f'ts = {timestamp-sesn.t0:.1f}')

        - To retrieve a list of all the recorded datasets named 'toto'

        >>> datas = [d for ts, d in sesn.datasets('toto')]

        """
        with self.conn as conn:
            c = conn.cursor()
            try:
                c.execute("SELECT name from dataset_names;")
                data = c.fetchall()
            except sqlite3.OperationalError:
                data = set()
            names = set([d[0] for d in data])
            if name not in names:
                print("Possible dataset names are", names)
                raise ValueError(f'Bad dataset name "{name:}"')
            it = c.execute(
                """SELECT timestamp, data FROM dataset
                              WHERE name='{:}'
                              ORDER BY timestamp ASC;
                           """.format(
                    name
                )
            )
            for row in it:
                yield row[0], pickle.loads(row[1])

    def dataset_last_data(self, name):
        """This method returns the last recorded dataset under the specified name.

        :param name: name of the dataset to retrieve
        :type name: str
        :return: dataset value
        :rtype: object
        """
        last_ts = self.dataset_times(name)[-1]
        return last_ts, self.dataset(name, ts=last_ts)

    def dataset_times(self, name):
        """This method returns the timestamp of the recorded dataset under the specified
        name.

        :param name: name of the dataset to retrieve
        :type name: str
        :return: array of timestamps
        :rtype: :class:`numpy.ndarray`
        """
        with self.conn as conn:
            c = conn.cursor()
            it = c.execute(
                """SELECT timestamp FROM dataset
                              WHERE name='{:}'
                              ORDER BY timestamp ASC;
                           """.format(
                    name
                )
            )
            t = np.array([v[0] for v in it])
        return t

    def dataset(self, name, ts=None, n=None):
        """This method returns the dataset recorded at the specified timestamp, and under
        the specified name.

        :param name: name of the dataset to retrieve
        :type name: str
        :param ts: timestamp at which the dataset was stored, defaults to the timestamp of the last recorded dataset under that name
        :type ts: float, optional
        :param n: select the nth dataset (in dataset timestamp chronological order)
        :type n: int, optional
        :return: the value of the recorded dataset
        :rtype: object
        """

        if ts is None:
            if n is None:
                ts, data = self.dataset_last_data(name)
                return data
            else:
                ts = self.dataset_times(name)[n]
        with self.conn as conn:
            c = conn.cursor()
            c.execute(
                """SELECT data FROM dataset
                         WHERE name='{:}' AND timestamp='{:}';
                      """.format(
                    name, ts
                )
            )
            data = pickle.loads(c.fetchone()[0])
        return data

    def save_parameter(self, *args, **kwargs):
        """This methods saves a scalar parameter into the database. Unlike scalar values
        saved by the :meth:`pymanip.asyncsession.AsyncSession.add_entry` method, such parameter
        can only hold one value, and does not have an associated timestamp.
        Parameters can be passed as dictionnaries, or keyword arguments.

        :param \\*args: dictionnaries with name-value to be added in the database
        :type \\*args: dict, optional
        :param \\**kwargs: name-value to be added in the database
        :type \\**kwargs: float, optional
        """
        data = dict()
        for a in args:
            data.update(a)
        data.update(kwargs)
        with self.conn as conn:
            c = conn.cursor()
            for key, val in data.items():
                c.execute(
                    """SELECT rowid FROM parameters
                             WHERE name='{:}';
                          """.format(
                        key
                    )
                )
                rowid = c.fetchone()
                if rowid is not None:
                    rowid = rowid[0]
                    c.execute(
                        """
                        REPLACE INTO parameters
                        (rowid, name, value)
                        VALUES (?,?,?);
                        """,
                        (rowid, key, val),
                    )
                else:
                    c.execute(
                        """
                        INSERT INTO parameters
                        (name, value)
                        VALUES (?,?);
                        """,
                        (key, val),
                    )

    def parameter(self, name):
        """This method retrieve the value of the specified parameter.

        :param name: name of the parameter to retrieve
        :type name: str
        :return: value of the parameter
        :rtype: float
        """
        with self.conn as conn:
            c = conn.cursor()
            c.execute(
                """
                      SELECT value FROM parameters
                      WHERE name='{:}';
                      """.format(
                    name
                )
            )
            data = c.fetchone()
            if data:
                return data[0]
        return None

    def has_parameter(self, name):
        """This method returns True if specified parameter exists in the session database.

        :param name: name of the parameter to retrieve
        :type name: str
        :return: True if parameter exists, False if it does not
        :rtype: bool
        """
        return self.parameter(name) is not None

    def parameters(self):
        """This method returns all parameter name and values.

        :return: parameters
        :rtype: dict
        """
        with self.conn as conn:
            c = conn.cursor()
            c.execute("SELECT * FROM parameters;")
            data = c.fetchall()
        return {d[0]: d[1] for d in data}

    def __getitem__(self, key):
        """Implement the evaluation of self[varname] as a shortcut to obtain timestamp and values for a given
        variable name.
        """
        return self.logged_variable(key)

    async def send_email(
        self,
        from_addr,
        to_addrs,
        host,
        port=None,
        subject=None,
        delay_hours=6,
        initial_delay_hours=None,
        use_ssl_submission=False,
        use_starttls=False,
        user=None,
        password=None,
    ):
        """This method returns an asynchronous task which sends an email at regular intervals.
        Such a task should be passed to :meth:`pymanip.asyncsession.AsyncSession.monitor` or
        :meth:`pymanip.asyncsession.AsyncSession.run`, and does not have to be awaited manually.

        :param from_addr: email address of the sender
        :type from_addr: str
        :param to_addrs: email addresses of the recipients
        :type to_addrs: list of str
        :param host: hostname of the SMTP server
        :type host: str
        :param port: port number of the SMTP server, defaults to 25
        :type port: int, optional
        :param delay_hours: interval between emails, defaults to 6 hours
        :type delay_hours: float, optional
        :param initial_delay_hours: time interval before the first email is sent, default to None (immediately)
        :type initial_delay_hours: float, optional
        """

        if port is None:
            if use_ssl_submission:
                port = 465
            elif use_starttls:
                # Tough some servers can do starttls on port 25, but then the user must specify port=25
                # as this is less common than on the submission port.
                port = 587
            else:
                port = 25
        if self.session_name is None:
            title = "Pymanip session"
        else:
            title = self.session_name
        if subject is None:
            subject = title

        if initial_delay_hours is None:
            initial_delay_hours = delay_hours / 2

        if initial_delay_hours > 0:
            await self.sleep(initial_delay_hours * 3600, verbose=False)

        jinja2_autoescape = jinja2.select_autoescape(["html"])
        jinja2_env = jinja2.Environment(
            loader=self.jinja2_loader, autoescape=jinja2_autoescape
        )
        template = jinja2_env.get_template("email.html")

        while self.running:

            dt_n = datetime.now()
            dt_fmt = "{:}{:02d}{:02d}-{:02d}{:02d}{:02d}"
            datestr = dt_fmt.format(
                dt_n.year, dt_n.month, dt_n.day, dt_n.hour, dt_n.minute, dt_n.second
            )
            # Generate HTML content
            last_values = self.logged_last_values()
            for name in last_values:
                timestamp, value = last_values[name]
                last_values[name] = (
                    timestamp,
                    value,
                    time.strftime(dateformat, time.localtime(timestamp)),
                )
            n_figs = len(self.figure_list)
            message_html = template.render(
                title=title,
                fignums=range(n_figs),
                datestr=datestr,
                last_values=last_values,
                variable_names=sorted(last_values.keys()),
            )

            # Create Email message
            msg = EmailMessage()
            msg["Subject"] = subject
            msg["From"] = from_addr
            msg["To"] = to_addrs
            msg.set_content("This is a MIME message")
            msg.add_alternative(message_html, subtype="html")

            # Add figure images
            for fignum, fig in enumerate(self.figure_list):
                fd, fname = tempfile.mkstemp(suffix=".png")
                with os.fdopen(fd, "wb") as f_png:
                    fig.canvas.draw_idle()
                    fig.savefig(f_png)
                with open(fname, "rb") as image_file:
                    figure_data = image_file.read()
                os.remove(fname)
                p = msg.get_payload()[1]
                p.add_related(
                    figure_data,
                    maintype="image",
                    subtype="png",
                    cid="{:d}{:}".format(fignum, datestr),
                    filename="fig{:d}-{:}.png".format(fignum, datestr),
                )

            try:
                if use_ssl_submission:
                    smtp_server = smtplib.SMTP_SSL(host, port)
                else:
                    smtp_server = smtplib.SMTP(host, port)
            except Exception as e:
                print("Unable to connect to SMTP server")
                print(e)
                await self.sleep(60, verbose=False)
                continue

            with smtp_server as smtp:
                if use_starttls:
                    smtp.starttls()
                if user and password:
                    if not use_starttls and not use_ssl_submission:
                        raise RuntimeError(
                            "Do you really want to send password unencrypted?"
                        )
                    try:
                        smtp.login(user, password)
                    except smtplib.SMTPHeloError:
                        print("The server didn’t reply properly to the HELO greeting.")
                    except smtplib.SMTPAuthenticationError:
                        print(
                            "The server didn’t accept the username/password combination."
                        )
                    except smtplib.SMTPNotSupportedError:
                        print("The AUTH command is not supported by the server.")
                try:
                    smtp.send_message(msg)
                    print("Email sent!")
                except smtplib.SMTPHeloError:
                    print("SMTP Helo Error")
                except smtplib.SMTPRecipientsRefused:
                    print("Some recipients have been rejected by SMTP server")
                except smtplib.SMTPSenderRefused:
                    print("SMTP server refused sender " + self.email_from_addr)
                except smtplib.SMTPDataError:
                    print("SMTP Data Error")

            await self.sleep(delay_hours * 3600, verbose=False)

    async def plot(
        self,
        varnames=None,
        maxvalues=1000,
        yscale=None,
        *,
        x=None,
        y=None,
        fixed_ylim=None,
        fixed_xlim=None,
    ):
        """This method returns an asynchronous task which creates and regularly updates a plot for
        the specified scalar variables. Such a task should be passed to :meth:`pymanip.asyncsession.AsyncSession.monitor` or
        :meth:`pymanip.asyncsession.AsyncSession.run`, and does not have to be awaited manually.

        If varnames is specified, the variables are plotted against time. If x and y are specified, then
        one is plotted against the other.

        :param varnames: names of the scalar variables to plot
        :type varnames: list or str, optional
        :param maxvalues: number of values to plot, defaults to 1000
        :type maxvalues: int, optional
        :param yscale: fixed yscale for temporal plot, defaults to automatic ylim
        :type yscale: tuple or list, optional
        :param x: name of the scalar variable to use on the x axis
        :type x: str, optional
        :param y: name of the scalar variable to use on the y axis
        :type y: str, optional
        :param fixed_xlim: fixed xscale for x-y plots, defaults to automatic xlim
        :type fixed_xlim: tuple or list, optional
        :param fixed_ylim: fixed yscale for x-y plots, defaults to automatic ylim
        :type fixed_ylim: tuple or list, optional
        """
        if varnames is None:
            if not isinstance(x, str) or not isinstance(y, str):
                raise TypeError("x and y should be strings")
            varnames = (x, y)
            param_key_window = "_window_xy_" + "_".join(varnames)
            param_key_figsize = "_figsize_xy_" + "_".join(varnames)
            xymode = True
        else:
            if x is not None or y is not None:
                raise ValueError("Cannot specify both varnames and (x,y)")
            if isinstance(varnames, str):
                varnames = (varnames,)
            param_key_window = "_window_" + "_".join(varnames)
            param_key_figsize = "_figsize_" + "_".join(varnames)
            xymode = False
        last_update = {k: 0 for k in varnames}
        saved_geom = self.parameter(param_key_window)
        if saved_geom:
            saved_geom = eval(saved_geom)
        saved_figsize = self.parameter(param_key_figsize)
        if saved_figsize:
            saved_figsize = eval(saved_figsize)
        if not self.offscreen_figures:
            plt.ion()
        fig = plt.figure(figsize=saved_figsize)
        if not self.offscreen_figures:
            mngr = fig.canvas.manager
            if saved_geom:
                mngr.window.setGeometry(saved_geom)
        ax = fig.add_subplot(111)
        line_objects = dict()
        self.figure_list.append(fig)
        ts0 = self.initial_timestamp
        while self.running:
            data = {
                k: self.logged_data_fromtimestamp(k, last_update[k]) for k in varnames
            }
            if xymode:
                ts_x, vs_x = data[x]
                ts_y, vs_y = data[y]
                if (ts_x != ts_y).any():
                    raise ValueError(
                        "xymode can only be used if x and y are synchronous"
                    )
                if ts_x.size > 0:
                    if y in line_objects:
                        p = line_objects[y]
                        xx = np.hstack((p.get_xdata(), vs_x))
                        yy = np.hstack((p.get_ydata(), vs_y))
                        p.set_xdata(xx)
                        p.set_ydata(yy)
                        if fixed_xlim is None:
                            xlim = ax.get_xlim()
                            try:
                                if xlim[1] < np.max(xx) or xlim[0] > np.min(xx):
                                    ax.set_xlim((np.min(xx), np.max(xx)))
                            except TypeError:
                                pass
                        if fixed_ylim is None:
                            ylim = ax.get_ylim()
                            try:
                                if ylim[1] < np.max(yy) or ylim[0] > np.min(yy):
                                    ax.set_ylim((np.min(yy), np.max(yy)))
                            except TypeError:
                                pass
                    else:
                        (p,) = ax.plot(vs_x, vs_y, "s-")
                        line_objects[y] = p
                        ax.set_xlabel(x)
                        ax.set_ylabel(y)
                        if fixed_xlim is None:
                            try:
                                if np.min(vs_x) != np.max(vs_x):
                                    ax.set_xlim((np.min(vs_x), np.max(vs_x)))
                            except TypeError:
                                pass
                        else:
                            ax.set_xlim(fixed_xlim)
                        if fixed_ylim is None:
                            try:
                                if np.min(vs_y) != np.max(vs_y):
                                    ax.set_ylim((np.min(vs_y), np.max(vs_y)))
                            except TypeError:
                                pass
                        else:
                            ax.set_ylim(fixed_ylim)
                        if not self.offscreen_figures:
                            fig.show()
                    last_update[x] = ts_x[-1]
                    last_update[y] = ts_y[-1]
            else:
                for name, values in data.items():
                    ts, vs = values
                    if ts.size > 0:
                        if name in line_objects:
                            # print('updating plot')
                            p = line_objects[name]
                            x = np.hstack((p.get_xdata(), (ts - ts0) / 3600))
                            y = np.hstack((p.get_ydata(), vs))
                            if x.size > maxvalues:
                                x = x[-maxvalues:]
                                y = y[-maxvalues:]
                            p.set_xdata(x)
                            p.set_ydata(y)
                            if x[0] != x[-1]:
                                ax.set_xlim((x[0], x[-1]))
                            if fixed_ylim is None:
                                ylim = ax.get_ylim()
                                if ylim[1] < np.max(y) or ylim[0] > np.min(y):
                                    ylim = (
                                        min((ylim[0], np.min(y))),
                                        max((ylim[1], np.max(y))),
                                    )
                                    ax.set_ylim(ylim)
                        else:
                            # print('initial plot')
                            x = (ts - ts0) / 3600
                            y = vs
                            if x.size > maxvalues:
                                x = x[-maxvalues:]
                                y = y[-maxvalues:]
                            (p,) = ax.plot(x, y, "o-", label=name)
                            line_objects[name] = p
                            ax.set_xlabel("t [h]")
                            if x[0] != x[-1]:
                                ax.set_xlim((x[0], x[-1]))
                            if yscale:
                                ax.set_yscale(yscale)
                            if fixed_ylim is not None:
                                ax.set_ylim(fixed_ylim)
                            ax.legend()
                            if not self.offscreen_figures:
                                fig.show()
                        last_update[name] = ts[-1]
            await asyncio.sleep(1)

        if not self.offscreen_figures:
            # Saving figure positions
            try:
                geom = mngr.window.geometry()
                figsize = tuple(fig.get_size_inches())
                self.save_parameter(
                    **{param_key_window: str(geom), param_key_figsize: str(figsize)}
                )
            except AttributeError:
                pass

    async def figure_gui_update(self):
        """This method returns an asynchronous task which updates the figures created by the
        :meth:`pymanip.asyncsession.AsyncSession.plot` tasks. This task is added automatically,
        and should not be used manually.
        """
        if not self.offscreen_figures:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=MatplotlibDeprecationWarning)
                while self.running:
                    figure_list = self.figure_list
                    if self.custom_figures:
                        figure_list = figure_list + self.custom_figures
                    if figure_list:
                        for fig in self.figure_list:
                            fig.canvas.start_event_loop(0.7 / len(self.figure_list))
                            await asyncio.sleep(0.3 / len(self.figure_list))
                        await asyncio.sleep(0.05)
                    else:
                        await asyncio.sleep(1.0)
        else:
            await asyncio.sleep(0.0)

    def ask_exit(self, *args, **kwargs):
        """This methods informs all tasks that the monitoring session should stop. Call this method if you
        wish to cleanly stop the monitoring session. It is also automatically called if the interrupt signal
        is received.
        It essentially sets the :attr:`running` attribute to False.
        Long user-defined tasks should check the :attr:`running` attribute, and abort if set to False.
        All other AsyncSession tasks check the attribute and will stop. The :meth:`~pymanip.asyncsession.AsyncSession.sleep`
        method also aborts sleeping.
        """
        self.running = False
        print(" Signal caught... stopping...")

    async def sweep(self, task, iterable):
        """This methods returns an asynchronous task which repeatedly awaits a given co-routine by iterating
        the specified iterable. The returned asynchronous task should be passed to :meth:`pymanip.asyncsession.AsyncSession.monitor` or
        :meth:`pymanip.asyncsession.AsyncSession.run`, and does not have to be awaited manually.

        This should be used when the main task of the asynchronous session is to sweep some value. The asynchronous session
        will exit when all values have been iterated.
        This is similar to running a script which consists in a synchronous for-loop, except that other tasks, such as remote
        access, live-plot and emails can be run concurrently.

        :param task: the co-routine function to repeatedly call and await
        :type task: function
        :param iterable: values to pass when calling the function
        :type iterable: list

        :Example:

        >>> async def balayage(sesn, voltage):
        >>>     await sesn.generator.vdc.aset(voltage)
        >>>     await asyncio.sleep(5)
        >>>     response = await sesn.multimeter.aget(channel)
        >>>     sesn.add_entry(voltage=voltage, response=response)
        >>>
        >>> async def main(sesn):
        >>>     await sesn.monitor(sesn.sweep(balayage, [0.0, 1.0, 2.0]))
        >>>
        >>> asyncio.run(main(sesn))

        """
        for val in iterable:
            await task(self, val)
            if not self.running:
                break
        self.running = False

    async def sleep(self, duration, verbose=True):
        """This method returns an asynchronous task which waits the specified amount of time and
        prints a countdown. This should be called with verbose=True by only one of the tasks.
        The other tasks should call with verbose=False.
        This method should be preferred over asyncio.sleep because it checks that
        :meth:`pymanip.asyncsession.AsyncSession.ask_exit` has not been called, and stops waiting
        if it has. This is useful to allow rapid abortion of the monitoring session.

        :param duration: time to wait
        :type duration: float
        :param verbose: whether to print the countdown, defaults to True
        :type verbose: bool, optional
        """
        start = time.monotonic()
        while self.running and time.monotonic() - start < duration:
            if verbose:
                print(
                    "Sleeping for "
                    + str(-int(time.monotonic() - start - duration))
                    + " s"
                    + " " * 8,
                    end="\r",
                )
                sys.stdout.flush()
            await asyncio.sleep(0.5)
        if verbose:
            sys.stdout.write("\n")

    async def server_main_page(self, request):
        """This asynchronous method returns the HTTP response to a request for the main html web page.
        Should not be called manually.
        """
        print("[", datetime.now(), request.remote, request.rel_url, "]")
        if self.session_name:
            context = {"title": self.session_name}
        else:
            context = {"title": "pymanip"}
        response = aiohttp_jinja2.render_template("main.html", request, context)
        return response

    async def server_logged_last_values(self, request):
        """This asynchronous method returns the HTTP response to a request for JSON data of the last logged
        values. Should not be called manually.
        """
        data = [
            {
                "name": name,
                "value": str(v[1]) if isinstance(v[1], bytes) else v[1],
                "datestr": time.strftime(dateformat, time.localtime(v[0])),
            }
            for name, v in self.logged_last_values().items()
        ]
        return web.json_response(data)

    async def server_get_parameters(self, request):
        """This asynchronous method returns the HTTP response to a request for JSON data of the session
        parameters. Should not be called manually.
        """
        params = {
            k: (str(v) if isinstance(v, bytes) else v)
            for k, v in self.parameters().items()
            if not k.startswith("_")
        }
        return web.json_response(params)

    async def server_plot_page(self, request):
        """This asynchronous method returns the HTTP response to a request for the HTML plot page.
        Should not be called manually.
        """
        print("[", datetime.now(), request.remote, request.rel_url, "]")
        context = {"name": request.match_info["name"]}
        response = aiohttp_jinja2.render_template("plot.html", request, context)
        return response

    async def server_data_from_ts(self, request):
        """This asynchronous method returns the HTTP response to a request for JSON with all data
        after the specified timestamp. Should not be called manually.
        """
        data_in = await request.json()
        last_ts = data_in["last_ts"]
        name = data_in["name"]
        timestamps, values = self.logged_data_fromtimestamp(name, last_ts)
        data_out = list(zip(timestamps, values))
        # print('from', last_ts, data_out)
        return web.json_response(data_out)

    async def server_current_ts(self, request):
        """This asynchronous method returns the HTTP response to a request for JSON with the current
        server time. Should not be called manually.
        """
        return web.json_response({"now": datetime.now().timestamp()})

    async def mytask(self, corofunc):
        """This method repeatedly awaits the given co-routine function, as long as
        :meth:`pymanip.asyncsession.AsyncSession.ask_exit` has not been called.
        Should not be called manually.
        """
        print("Starting task", corofunc)
        sig = inspect.signature(corofunc)
        while self.running:
            if len(sig.parameters) == 1:
                await corofunc(self)
            else:
                await corofunc()
        print("Task finished", corofunc)

    async def monitor(
        self,
        *tasks,
        server_port=6913,
        custom_routes=None,
        custom_figures=None,
        offscreen_figures=False,
    ):
        """This method runs the specified tasks, and opens a web-server for remote access and set up the tasks to run
        matplotlib event loops if necessary. This is the main
        method that the main function of user program should await for. It is also responsible for setting up
        the signal handling and binding it to the ask_exit method.

        It defines a :attr:`running` attribute, which is finally set to False when the monitoring must stop. User can
        use the :meth:`~pymanip.asyncsession.AsyncSession.ask_exit` method to stop the monitoring. Time consuming
        user-defined task should check the :attr:`running` and abort if set to False.

        :param \\*tasks: asynchronous tasks to run: if the task is a co-routine function, it will be called repeatedly until ask_exit is called. If task is an awaitable it is called only once. Such an awaitable is responsible to check that ask_exit has not been called. Several such awaitables are provided: :meth:`pymanip.asyncsession.AsyncSession.send_email`, :meth:`pymanip.asyncsession.AsyncSession.plot` and :meth:`pymanip.asyncsession.AsyncSession.sweep`.
        :type \\*tasks: co-routine function or awaitable
        :param server_port: the network port to open for remote HTTP connection, defaults to 6913. If None, no server is created.
        :type server_port: int, optional
        :param custom_routes: additionnal aiohttp routes for the web-server, defaults to None
        :type custom_routes: co-routine function, optional
        :param custom_figures: additional matplotlib figure object that needs to run the matplotlib event loop
        :type custom_figures: :class:`matplotlib.pyplot.Figure`, optional
        :param offscreen_figures: if set, figures are not shown onscreen
        :type offscreen_figures: bool, optional
        """
        loop = asyncio.get_event_loop()
        self.custom_figures = custom_figures
        self.offscreen_figures = offscreen_figures

        # signal handling
        self.running = True
        if sys.platform == "win32":
            # loop.add_signal_handler raises NotImplementedError
            signal.signal(signal.SIGINT, self.ask_exit)
        else:
            for signame in ("SIGINT", "SIGTERM"):
                loop.add_signal_handler(getattr(signal, signame), self.ask_exit)

        # web server
        if server_port:
            app = web.Application(loop=loop)
            aiohttp_jinja2.setup(app, loader=self.jinja2_loader)
            app.router.add_routes(
                [
                    web.get("/", self.server_main_page),
                    web.get("/api/logged_last_values", self.server_logged_last_values),
                    web.get("/plot/{name}", self.server_plot_page),
                    web.static("/static", self.static_dir),
                    web.post("/api/data_from_ts", self.server_data_from_ts),
                    web.get("/api/server_current_ts", self.server_current_ts),
                    web.get("/api/get_parameters", self.server_get_parameters),
                ]
            )
            if custom_routes:
                app.router.add_routes(custom_routes)

            webserver = loop.create_server(
                app.make_handler(), host=None, port=server_port
            )

        # if any of the tasks submitted are coroutinefunctions instead of
        # coroutines, then assume they take only one argument (self)
        tasks_final = list()
        for t in tasks:
            if asyncio.iscoroutinefunction(t):
                tasks_final.append(self.mytask(t))
            elif asyncio.iscoroutine(t):
                tasks_final.append(t)
            else:
                raise TypeError("Coroutine or Coroutinefunction is expected")
        print("Starting event loop")
        if server_port:
            await asyncio.gather(webserver, self.figure_gui_update(), *tasks_final)
        else:
            await asyncio.gather(self.figure_gui_update(), *tasks_final)

    def run(
        self,
        *tasks,
        server_port=6913,
        custom_routes=None,
        custom_figures=None,
        offscreen_figures=False,
    ):
        """Synchronous call to :meth:`pymanip.asyncsession.AsyncSession.monitor`."""

        asyncio.run(
            self.monitor(
                *tasks,
                server_port=server_port,
                custom_routes=custom_routes,
                custom_figures=custom_figures,
                offscreen_figures=offscreen_figures,
            )
        )

    def save_remote_data(self, data):
        """This method saves the data returned by a :class:`pymanip.asyncsession.RemoteObserver` object into the current session database,
        as datasets and parameters.

        :param data: data returned by the :class:`pymanip.asyncsession.RemoteObserver` object
        :type data: dict
        """
        for k, v in data.items():
            # print(k,type(v),v)
            try:
                v[0]
                iterable = True
            except (TypeError, KeyError):
                iterable = False
            if iterable:
                # we are iterable
                self.add_dataset(**{k: v})
            else:
                # we are not iterable
                if isinstance(v, dict):
                    # non reduced data, v is a dictionnary with two keys, 't' and 'value'
                    self.add_dataset(**{k: v["value"]})
                    self.add_dataset(**{k + "_time": v["t"]})
                else:
                    try:
                        # data must be a scalar
                        float(v)
                    except TypeError:
                        print("skipping", k, type(v))
                        continue
                    self.save_parameter(**{k: v})


class RemoteObserver:
    """This class represents remote observers of a monitoring session. It connects to the server opened on a remote computer by
    :meth:`pymanip.asyncsession.AsyncSession.monitor`. The aim of an instance of RemoteObserver is to retrieve the data saved into
    the remote computer session database.

    :param host: hostname of the remote compute to connect to
    :type host: str
    :param port: port number to connect to, defaults to 6913
    :type port: int, optional
    """

    def __init__(self, host, port=6913):
        """Constructor method"""
        self.host = host
        self.port = port

    def _get_request(self, apiname):
        """Private method to send a GET request for the specified API name"""
        url = "http://{host:}:{port:}/api/{api:}".format(
            host=self.host, port=self.port, api=apiname
        )
        r = requests.get(url)
        try:
            return r.json()
        except json.decoder.JSONDecodeError:
            print(r.text)
            raise

    def _post_request(self, apiname, params):
        """Private method to send a POST request for the specified API name and params"""
        url = "http://{host:}:{port:}/api/{api:}".format(
            host=self.host, port=self.port, api=apiname
        )
        r = requests.post(url, json=params)
        try:
            return r.json()
        except json.decoder.JSONDecodeError:
            print(r.text)
            raise

    def get_last_values(self):
        """This method retrieve the last set of values from
        the remote monitoring session.

        :return: scalar variable last recorded values
        :rtype: dict
        """

        data = self._get_request("logged_last_values")
        return {d["name"]: d["value"] for d in data}

    def start_recording(self):
        """This method establishes the connection to the remote computer, and sets the
        start time for the current observation session.
        """
        self.server_ts_start = self._get_request("server_current_ts")["now"]
        data = self.get_last_values()
        self.remote_varnames = list(data.keys())

    def stop_recording(self, reduce_time=True, force_reduce_time=True):
        """This method retrieves all scalar variable data recorded saved by the remote
        computer since :meth:`pymanip.asyncsession.RemoteObserver.start_recording` established
        the connection.

        :param reduce_time: if True, try to collapse all timestamp arrays into a unique timestamp array. This is useful if the remote computer program only has one call to add_entry. Defaults to True.
        :type reduce_time: bool, optional
        :param force_reduce_time: bypass checks that all scalar values indeed have the same timestamps.
        :type force_reduce_time: bool, optional
        :return: timestamps and values of all data saved in the remote computed database since the call to :meth:`pymanip.asyncsession.RemoteObserver.start_recording`
        :rtype: dict
        """
        recordings = dict()
        for varname in self.remote_varnames:
            data = self._post_request(
                "data_from_ts",
                params={"name": varname, "last_ts": self.server_ts_start},
            )
            if len(data) > 0:
                recordings[varname] = {
                    "t": [d[0] for d in data],
                    "value": [d[1] for d in data],
                }
        if reduce_time:
            t = recordings[self.remote_varnames[0]]["t"]
            if (
                all([recordings[varname]["t"] == t for varname in recordings])
                or force_reduce_time
            ):
                recordings = {k: v["value"] for k, v in recordings.items()}
                recordings["time"] = t
            else:
                print("t =", t)
                pprint(
                    {
                        varname: recordings[varname]["t"] == t
                        for varname in self.remote_varnames
                    }
                )
        parameters = self._get_request("get_parameters")
        recordings.update(parameters)

        return recordings


class SavedAsyncSession:
    """This class implements the same methods as AsyncSession with readonly mode, but with
    caching enabled. Also, the file is opened on demand. No context manager is necessary.
    """

    def __init__(self, session_name, verbose=True):

        self.session_name = session_name
        self.verbose = verbose
        self.session = AsyncSession(session_name, verbose=False, readonly=True)
        if verbose:
            try:
                self.print_welcome()
            except sqlite3.OperationalError:
                print("Warning: unable to open database for session", session_name)

    def print_welcome(self):
        """Prints informative start date/end date message. If verbose is True, this method
        is called by the constructor.
        """
        start_string = time.strftime(dateformat, time.localtime(self.initial_timestamp))
        cprint.blue("*** Start date: " + start_string)
        last = self.last_timestamp
        if last:
            end_string = time.strftime(dateformat, time.localtime(last))
            cprint.blue("***   End date: " + end_string)

    def print_description(self):
        """Prints the list of parameters, logged variables and datasets."""
        version = self.get_version()
        print(
            self.session_name,
            "is an asynchroneous session (version {:}).".format(version),
        )
        print()
        last_values = self.logged_last_values()
        params = {
            key: val
            for key, val in self.parameters().items()
            if not key.startswith("_")
        }
        if params:
            print("Parameters")
            print("==========")
            for key, val in self.parameters().items():
                print(key, ":", val)
            print()

        if last_values:
            print("Logged variables")
            print("================")
            for name, t_v in last_values.items():
                print(name, "(", t_v[1], ")")
            print()

        ds_names = self.dataset_names()
        if ds_names:
            print("Datasets")
            print("========")
            for ds in ds_names:
                print(ds)
            print()

    # General attributes

    @lru_cache(maxsize=128)
    def get_version(self):
        with self.session as sesn:
            return sesn.get_version()

    @property
    def t0(self):
        with self.session as sesn:
            return sesn.t0

    @property
    def initial_timestamp(self):
        return self.t0

    @cached_property
    def last_timestamp(self):
        ts = list()
        last_values = self.logged_last_values()
        if last_values:
            ts.append(max([t_v[0] for name, t_v in last_values.items()]))
        for ds_name in self.dataset_names():
            ts.append(max(self.dataset_times(ds_name)))
        if ts:
            return max(ts)
        return None

    # Logged variables

    @lru_cache(maxsize=128)
    def logged_data(self):
        with self.session as sesn:
            return sesn.logged_data()

    def logged_variables(self):
        return set(self.logged_data().keys())

    def logged_variable(self, varname):
        return self.logged_data()[varname]

    @lru_cache(maxsize=128)
    def logged_last_values(self):
        last = dict()
        for name, data in self.logged_data().items():
            ts, val = data
            max_ts = np.max(ts)
            assert ts[-1] == max_ts
            last[name] = (ts[-1], val[-1])
        return last

    @lru_cache(maxsize=128)
    def logged_first_values(self):
        first = dict()
        for name, data in self.logged_data().items():
            ts, val = data
            min_ts = np.min(ts)
            assert ts[0] == min_ts
            first[name] = (ts[0], val[0])
        return first

    @lru_cache(maxsize=128)
    def logged_data_fromtimestamp(self, name, timestamp):
        ts, val = self.logged_variable(name)
        for ind, tt in enumerate(ts):
            if tt == timestamp:
                break
        else:
            raise ValueError("No logged data at specified timestamp")
        return val[ind]

    def __getitem__(self, key):
        return self.logged_variable(key)

    # Dataset

    @lru_cache(maxsize=128)
    def dataset_names(self):
        with self.session as sesn:
            return sesn.dataset_names()

    def datasets(self, name):
        with self.session as sesn:
            return sesn.datasets()

    def dataset_last_data(self, name):
        """This method returns the last recorded dataset under the specified name.

        :param name: name of the dataset to retrieve
        :type name: str
        :return: dataset value
        :rtype: object
        """
        last_ts = self.dataset_times(name)[-1]
        return last_ts, self.dataset(name, ts=last_ts)

    @lru_cache(maxsize=128)
    def dataset_times(self, name):
        with self.session as sesn:
            sesn.dataset_times(name)

    def dataset(self, name, ts=None, n=None):
        if ts is None:
            if n is None:
                ts, data = self.dataset_last_data(name)
                return data
            else:
                ts = self.dataset_times(name)[n]
        with self.session as sesn:
            return sesn.dataset(name, ts=ts, n=n)

    # Parameters

    @lru_cache(maxsize=128)
    def parameters(self):
        with self.session as sesn:
            return sesn.parameters()

    def parameter(self, name):
        return self.parameters()[name]

    def has_parameter(self, name):
        return name in self.parameters()


if __name__ == "__main__":
    with AsyncSession("Essai") as sesn:
        sesn.add_entry(a=1, b=2)
        sesn.save_parameter(c=3)
        sesn.plot("a")



"""AVT Camera module (:mod:`pymanip.video.avt`)
===============================================

This module implements the :class:`pymanip.video.avt.AVT_Camera` class
using the third-party :mod:`pymba` module.

.. autoclass:: AVT_Camera
   :members:
   :private-members:
   :show-inheritance:

"""

import numpy as np
from pymanip.video import MetadataArray, Camera, CameraTimeout
from pymanip.asynctools import synchronize_generator
from pymba import Vimba
from pymba.vimbaexception import VimbaException
import sys


class AVT_Camera(Camera):
    """Concrete :class:`pymanip.video.Camera` class for AVT camera.

    :param cam_num: camera number, defaults to 0
    :type cam_num: int, optional
    :param pixelFormat: pixel format to use, e.g. "Mono8", "Mono16". Defaults to "Mono16"
    :type pixelFormat: str, optional

    """

    # Class attributes
    system = None
    vimba = None
    active_cameras = set()

    @classmethod
    def get_camera_list(cls):
        """This methods returns the list of camera numbers connected to
        the computer.

        :return: camera ids
        :rtype: list
        """
        if cls.vimba:
            return cls.vimba.getCameraIds()
        else:
            with Vimba() as vimba_:
                return vimba_.getCameraIds()

    def __init__(self, cam_num=0, pixelFormat="Mono16"):
        """Constructor method
        """
        if not AVT_Camera.vimba:
            AVT_Camera.vimba = Vimba().__enter__()
        if not AVT_Camera.system:
            AVT_Camera.system = AVT_Camera.vimba.getSystem()
        self.cameraDesc = AVT_Camera.get_camera_list()[cam_num]
        # self.info = vimba.getCameraInfo(self.cameraDesc)
        self.camera = AVT_Camera.vimba.getCamera(self.cameraDesc)
        self.camera.openCamera()
        AVT_Camera.active_cameras.add(self)
        self.num = cam_num
        self.name = "AVT " + str(cam_num)
        self.model_name = self.camera.DeviceModelName.decode("ascii")
        print("cam" + str(self.num) + ":", self.model_name)
        # par défaut, on démarre en Mode 0 avec la résolution maximale à la
        # création de l'objet
        self.camera.IIDCMode = "Mode0"
        # défaut: ROI entier
        self.camera.OffsetX = 0
        self.camera.OffsetY = 0
        self.camera.Width = self.camera.WidthMax
        self.camera.Height = self.camera.HeightMax
        print(
            "cam" + str(self.num) + ": maximum resolution is",
            self.camera.WidthMax,
            "x",
            self.camera.HeightMax,
        )
        print(
            "cam" + str(self.num) + ": maximum framerate at full resolution is",
            self.camera_feature_info("AcquisitionFrameRate")["range"][1],
        )
        self.camera.PixelFormat = pixelFormat.encode("ascii")

    def close(self):
        """This method closes the connection to the camera.
        """
        self.camera.closeCamera()
        self.camera = None
        AVT_Camera.active_cameras.remove(self)
        if len(AVT_Camera.active_cameras) == 0:
            AVT_Camera.system = None
            AVT_Camera.vimba.__exit__(None, None, None)
            AVT_Camera.vimba = None

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        super(AVT_Camera, self).__exit__(type_, value, cb)
        self.close()

    def camera_features(self):
        """This methods returns the list of possible camera features.

        :return: camera feature
        :rtype: list
        """
        cameraFeatureNames = self.camera.getFeatureNames()
        return [f.decode("ascii") for f in cameraFeatureNames]

    def camera_feature_info(self, featureName):
        """This method queries the camera for the specified feature.

        :param featureName: one of the features returned by :meth:`~pymanip.video.avt.AVT_Camera.camera_features`
        :type featureName: str
        :return: values associated with the specified feature
        :rtype: dict
        """
        featureInfo = self.camera.getFeatureInfo(featureName)
        featDict = {
            field.decode("ascii")
            if isinstance(field, bytes)
            else field: getattr(featureInfo, field)
            for field in featureInfo.getFieldNames()
        }
        featDict["value"] = getattr(self.camera, featureName)
        featDict["range"] = self.camera.getFeatureRange(featureName)
        featDict["featureDataTypeName"] = {
            0: "Unknown feature type",
            1: "64 bit integer",
            2: "64 bit floating point",
            3: "Enumeration feature",
            4: "String feature",
            5: "Boolean feature",
            6: "Command feature",
            7: "Raw (direct register access) feature",
            8: "Feature with no data",
        }[featDict["featureDataType"]]
        featDict["featureFlagsStr"] = ""
        if featDict["featureFlags"] & 1:
            featDict["featureFlagsStr"] += "r"  # read access
        if featDict["featureFlags"] & 2:
            featDict["featureFlagsStr"] += "w"  # write access
        if featDict["featureFlags"] & 8:
            featDict["featureFlagsStr"] += "v"  # volatile
        if featDict["featureFlags"] & 16:
            featDict["featureFlagsStr"] += "m"  # may change after write
        for k in featDict:
            if isinstance(featDict[k], bytes):
                featDict[k] = featDict[k].decode("ascii")
        return featDict

    def acquisition_oneshot(self):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_oneshot` for the AVT camera.
        """
        self.camera.AcquisitionMode = "SingleFrame"
        self.frame = self.camera.getFrame()
        self.frame.announceFrame()
        self.camera.startCapture()
        try:
            self.frame.queueFrameCapture()
            self.camera.runFeatureCommand("AcquisitionStart")
            self.camera.runFeatureCommand("AcquisitionStop")
            self.frame.waitFrameCapture()
            # print('timestamp =', self.frame.timestamp)
            # print('pixel_bytes =', self.frame.pixel_bytes)
            if self.frame.pixel_bytes == 1:
                dt = np.uint8
            elif self.frame.pixel_bytes == 2:
                dt = np.uint16
            else:
                raise NotImplementedError
            img = MetadataArray(
                np.ndarray(
                    buffer=self.frame.getBufferByteData(),
                    dtype=dt,
                    shape=(self.frame.height, self.frame.width),
                ).copy(),
                metadata={"timestamp": self.frame.timestamp * 1e-7},
            )

        finally:
            self.camera.endCapture()
            self.camera.revokeAllFrames()

        return img

    def set_trigger_mode(self, mode=False):
        """This method sets the trigger mode for the camera.
        For external trigger, the method also sets IIDCPacketSizeAuto to 'Maximize'.

        :param mode: True if external trigger. False if internal trigger.
        :type mode: bool
        """

        if mode:
            self.camera.TriggerMode = "On"
            self.camera.TriggerSource = "InputLines"
            self.camera.IIDCPacketSizeAuto = "Maximize"
        else:
            self.camera.TriggerMode = "Off"

    def set_exposure_time(self, seconds):
        """This method sets the exposure time for the camera.

        :param seconds: exposure in seconds. Possible values range from 33.0 µs to 67108895.0 µs.
        :type seconds: float

        """
        self.camera.ExposureMode = "Timed"
        self.camera.ExposureTime = seconds * 1e6

    def set_roi(self, roiX0=0, roiY0=0, roiX1=0, roiY1=0):
        """This mothods sets the position of the upper left corner (X0, Y0) and the lower right (X1, Y1)
        corner of the ROI (region of interest) in pixels.
        """
        if (roiX1 - roiX0) <= 0 or (roiX1 - roiX0) > self.camera.WidthMax:
            raise ValueError("Width must be > 0 and < WidthMax")
        if roiX0 < 0:
            raise ValueError("X0 must be >= 0")
        assert roiX1 >= 0
        if (roiY1 - roiY0) <= 0 or (roiY1 - roiY0) > self.camera.HeightMax:
            raise ValueError("Height must be > 0 and < HeightMax")
        if roiY0 < 0:
            raise ValueError("Y0 must be positive")
        assert roiY1 >= 0

        # On met d'abord l'offset à 0 pour être sûr que la largeur et hauteur est acceptable
        self.camera.OffsetX = 0
        self.camera.OffsetY = 0
        # On change d'abord la largeur et hauteur pour que l'offset soit acceptable
        self.camera.Width = roiX1 - roiX0
        self.camera.Height = roiY1 - roiY0
        # Dernière étape...
        self.camera.OffsetX = roiX0
        self.camera.OffsetY = roiY0

        print(
            "cam" + str(self.num) + ": resolution is",
            self.camera.Width,
            "x",
            self.camera.Height,
        )
        print(
            "cam" + str(self.num) + ": maximum framerate is",
            self.camera_feature_info("AcquisitionFrameRate")["range"][1],
        )

    def acquisition(
        self,
        num=np.inf,
        timeout=1000,
        raw=False,
        framerate=None,
        external_trigger=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition` for the AVT camera.
        """
        yield from synchronize_generator(
            self.acquisition_async,
            num=num,
            timeout=timeout,
            raw=raw,
            framerate=framerate,
            external_trigger=external_trigger,
            initialising_cams=initialising_cams,
            raise_on_timeout=raise_on_timeout,
        )

    async def acquisition_async(
        self,
        num=np.inf,
        timeout=1000,
        raw=False,
        framerate=None,
        external_trigger=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_async` for the AVT camera.
        """
        self.camera.AcquisitionMode = "Continuous"
        if framerate is not None:
            # Not usable if HighSNRIImages>0, external triggering or
            # IIDCPacketSizeAuto are active
            self.camera.AcquisitionFrameRate = framerate
        if external_trigger:
            self.camera.TriggerMode = "On"
            self.camera.TriggerSource = "InputLines"
        self.frame = self.camera.getFrame()
        self.frame.announceFrame()
        self.camera.startCapture()
        self.camera.runFeatureCommand("AcquisitionStart")
        self.buffer_queued = False
        try:
            count = 0
            while count < num:
                if not self.buffer_queued:
                    self.frame.queueFrameCapture()
                    self.buffer_queued = True
                if (
                    count == 0
                    and initialising_cams is not None
                    and self in initialising_cams
                ):
                    initialising_cams.remove(self)
                errorCode = await self.frame.waitFrameCapture_async(int(timeout))
                if errorCode == -12:
                    if raise_on_timeout:
                        raise CameraTimeout("cam" + str(self.num) + " timeout")
                    else:
                        stop_signal = yield None
                        if stop_signal:
                            break
                        else:
                            continue
                elif errorCode != 0:
                    raise VimbaException(errorCode)
                if self.frame.pixel_bytes == 1:
                    dt = np.uint8
                elif self.frame.pixel_bytes == 2:
                    dt = np.uint16
                else:
                    raise NotImplementedError
                self.buffer_queued = False
                stop_signal = yield MetadataArray(
                    np.ndarray(
                        buffer=self.frame.getBufferByteData(),
                        dtype=dt,
                        shape=(self.frame.height, self.frame.width),
                    ),
                    metadata={
                        "counter": count,
                        "timestamp": self.frame.timestamp * 1e-7,
                    },
                )
                count += 1
                if stop_signal:
                    break

        finally:
            self.camera.runFeatureCommand("AcquisitionStop")
            self.camera.endCapture()
            self.camera.revokeAllFrames()
        if stop_signal:
            yield True


if __name__ == "__main__":
    try:
        output = open(sys.argv[1], "w")
    except IndexError:
        output = sys.stdout

    list = AVT_Camera.get_camera_list()
    for l in list:
        print(l.decode("ascii"), file=output)

    with AVT_Camera(0) as cam:
        # features
        for f in cam.camera_features():
            print(f, file=output)

        print("-" * 8, file=output)

        img = cam.acquisition_oneshot()

        for feat in [
            "PixelFormat",
            "AcquisitionFrameRate",
            "AcquisitionMode",
            "ExposureAuto",
            "ExposureMode",
            "ExposureTime",
            "IIDCMode",
            "IIDCIsoChannel",
            "IIDCPacketSize",
            "IIDCPacketSizeAuto",
            # 'IIDCPhyspeed', 'TriggerMode',
            # 'TriggerSource', 'TriggerDelay', 'Gain', 'GainAuto', 'GainSelector',
            # 'Width', 'WidthMax', 'Height', 'HeightMax', 'HighSNRImages'
        ]:
            featDict = cam.camera_feature_info(feat)
            for k, v in featDict.items():
                print(k, ":", v, file=output)
            print("-" * 8, file=output)

    import matplotlib.pyplot as plt

    plt.imshow(img, origin="lower", cmap="gray")
    plt.show()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import numpy as np


def load_octmi_dat(acquisitionName, basePath="."):
    # Vérification de l'existence du fichier
    datFilePath = os.path.join(os.path.normpath(basePath), acquisitionName + "_MI.dat")
    if not os.path.exists(datFilePath):
        print("Could not stat file", datFilePath)
        raise NameError("File does not exist")

    # Décompte du nombre d'éléments
    nval = 0
    variableList = ""
    with open(datFilePath, "r") as f:
        for line in f:
            if line[0] == "T":
                if line != variableList:
                    variableList = line
                    # print variableList
            else:
                nval = nval + 1

    variableList = variableList.split(" ")

    dictionnaire = dict()
    dictionnaire["nval"] = nval
    if nval > 1:
        for i in range(len(variableList)):
            dictionnaire[variableList[i].strip()] = np.zeros(nval)

    linenum = 0
    with open(datFilePath, "r") as f:
        for line in f:
            contentList = line.split(" ")
            if contentList[0] != "Time":
                if nval == 1:
                    for i in range(len(variableList)):
                        dictionnaire[variableList[i].strip()] = eval(
                            contentList[i].strip()
                        )
                else:
                    for i in range(len(variableList)):
                        if i < len(contentList):
                            dataStr = contentList[i].strip()
                            if dataStr.lower() == "nan":
                                dictionnaire[variableList[i].strip()][linenum] = np.nan
                            else:
                                dictionnaire[variableList[i].strip()][linenum] = eval(
                                    contentList[i].strip()
                                )
                        else:
                            dictionnaire[variableList[i].strip()][linenum] = np.nan
                    linenum = linenum + 1

    return dictionnaire



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""

This module contains all the instrument classes available in fluidlab.

List of supported instruments:

"""

from __future__ import print_function
import sys
import fluidlab.instruments as instruments
import pymanip

if hasattr(pymanip, "pymanip_import_verbose"):
    verbose_import = pymanip.pymanip_import_verbose
else:
    verbose_import = False

__all__ = []

if verbose_import:
    print("Supported instruments:")

for instrument_type in instruments.__all__:
    exec("import fluidlab.instruments." + instrument_type + " as " + instrument_type)
    exec("instrument_type_module = " + instrument_type)
    __doc__ += "* " + instrument_type + ": "
    if verbose_import:
        sys.stdout.write("* " + instrument_type + ": ")
    for instrument_file in instrument_type_module.__all__:  # noqa: F821
        try:
            exec(
                "import fluidlab.instruments."
                + instrument_type
                + "."
                + instrument_file
                + " as "
                + instrument_file
            )
            exec("instrument_file_module = " + instrument_file)
            for instrument_classname in instrument_file_module.__all__:  # noqa: F821
                exec(
                    "from fluidlab.instruments."
                    + instrument_type
                    + "."
                    + instrument_file
                    + " import "
                    + instrument_classname
                )
                __doc__ += instrument_classname + " "
                if verbose_import:
                    sys.stdout.write(instrument_classname + " ")
                __all__.append(instrument_classname)
        except ImportError:
            print(
                "Unable to import",
                instrument_type + "." + instrument_file + "." + instrument_classname,
            )

    __doc__ += "\n"
    if verbose_import:
        sys.stdout.write("\n")

__doc__ += """
Example of usage:

from pymanip.instruments import Agilent34401a
from pymanip.interfaces import GPIBInterface

multiplexer = Agilent34401a(GPIBInterface(0,5))
"""



"""Concrete implementation with niscope (:mod:`pymanip.aiodaq.scope`)
=====================================================================

This module is a concrete implementation of the :class:`~pymanip.aiodaq.AcquisitionCard`
class using the :mod:`niscope` module.

.. note::
   We have tested this module only for with a PXI-5922 card.

.. autoclass:: ScopeSystem
   :members:
   :private-members:

.. autofunction:: get_device_list

"""

import asyncio
import time
import warnings
import subprocess as sp

from niScope import Scope, SLOPE, TRIGGER_SOURCE, ScopeException

from pymanip.aiodaq import TriggerConfig, AcquisitionCard, TimeoutException

try:
    from pymanip.aiodaq.daqmx import get_device_list as daqmx_get_devices

    has_daqmx = True
except ImportError:
    has_daqmx = False
try:
    from pymanip.nisyscfg import scope_devices

    has_nisyscfg = True
except (ImportError, OSError):
    has_nisyscfg = False

possible_sample_rates = [60e6 / n for n in range(4, 1201)]


class ScopeSystem(AcquisitionCard):
    """This class is the concrete implentation for NI Scope cards.

    :param scope_name: the name of the scope device, e.g. "Dev1"
    :type scope_name: str

    """

    def __init__(self, scope_name=None):
        super(ScopeSystem, self).__init__()
        self.scope_name = scope_name
        if scope_name:
            self.scope = Scope(scope_name)
        else:
            self.scope = None
        self.trigger_set = False

    @property
    def samp_clk_max_rate(self):
        """Maximum rate for the board clock.
        """
        return max(possible_sample_rates)

    def possible_trigger_channels(self):
        """This method returns the list of possible channels for external triggering.
        """
        return ["Ext"] + self.channels

    def close(self):
        """This method closes the connection to the board.
        """
        if self.scope:
            self.scope.close()
            self.scope = None

    def add_channel(self, channel_name, terminal_config, voltage_range):
        """Concrete implementation of :meth:`pymanip.aiodaq.AcquisitionCard.add_channel`.
        """
        cn = str(channel_name)
        if "/" in cn:
            scope_name, cn = cn.split("/")
        else:
            scope_name = self.scope_name
        if not self.scope:
            self.scope_name = scope_name
            self.scope = Scope(scope_name)
        if cn not in self.channels:
            self.channels.append(cn)
            self.scope.ConfigureVertical(channelList=cn, voltageRange=voltage_range)
            actual_range = self.scope.ActualVoltageRange(cn)
            self.actual_ranges.append(actual_range)
            if actual_range != voltage_range:
                warnings.warn(
                    f"Actual range is {actual_range:} " f"for chan {cn:}.",
                    RuntimeWarning,
                )

    def configure_clock(self, sample_rate, samples_per_chan):
        """Concrete implementation for :meth:`pymanip.aiodaq.AcquisitionCard.configure_clock`
        """
        if sample_rate not in possible_sample_rates:
            initial_sample_rate = sample_rate
            chosen = possible_sample_rates[0]
            for pos in possible_sample_rates:
                if abs(sample_rate - pos) < abs(chosen - pos):
                    chosen = pos
            sample_rate = chosen
            warnings.warn(
                f"{initial_sample_rate:} is not possible. "
                f"Closest sample rate is {sample_rate:}",
                RuntimeWarning,
            )
        self.scope.ConfigureHorizontalTiming(
            sampleRate=sample_rate, numPts=samples_per_chan
        )
        self.scope.NumRecords = 1
        self.sample_rate = self.scope.ActualSamplingRate
        self.samples_per_chan = self.scope.ActualRecordLength
        print("sample_rate =", self.sample_rate)

    def configure_trigger(
        self,
        trigger_source=None,
        trigger_level=0,
        trigger_config=TriggerConfig.EdgeRising,
    ):
        """Concrete implementation for :meth:`pymanip.aiodaq.AcquisitionCard.configure_trigger`
        """
        if trigger_source is None:
            print("Setting trigger to Immediate")
            self.scope.ConfigureTrigger("Immediate")
        else:
            trigger_source = str(trigger_source)
            if "/" in trigger_source:
                scope_name, trigger_source = trigger_source.split("/")
                if scope_name != self.scope_name:
                    raise ValueError("Wrong trigger source")
            print(
                "Setting trigger_source to", trigger_source, "level to", trigger_level
            )
            if trigger_source == "Ext":
                trigger_source = TRIGGER_SOURCE.EXTERNAL
            try:
                trigger_source = trigger_source.encode("ascii")
            except AttributeError:
                pass
            self.scope.ConfigureTrigger(
                "Edge",
                triggerSource=trigger_source,
                slope=SLOPE.POSITIVE,
                level=float(trigger_level),
            )
        self.trigger_set = True

    def start(self):
        """Concrete implementation for :meth:`pymanip.aiodaq.AcquisitionCard.start`
        """
        if not self.trigger_set:
            self.configure_trigger()
        self.scope.InitiateAcquisition()
        self.running = True

    async def stop(self):
        """Concrete implementation for :meth:`pymanip.aiodaq.AcquisitionCard.stop`
        """
        if self.running:
            self.running = False
            while self.reading:
                await asyncio.sleep(1.0)
            self.scope.Abort()

    async def read(self, tmo=None):
        """Concrete implementation for :meth:`pymanip.aiodaq.AcquisitionCard.read`
        """
        self.reading = True
        start = time.monotonic()
        data = None
        while self.running:
            try:
                data = await self.loop.run_in_executor(
                    None, self.scope.Fetch, ",".join(self.channels), None, 1.0
                )
                break
            except ScopeException:
                if tmo and time.monotonic() - start > tmo:
                    raise TimeoutException()
                else:
                    continue
        if data is not None:
            self.last_read = time.monotonic()
            if len(self.channels) > 1:
                data = data.T
            else:
                data = data[:, 0]
        self.reading = False
        return data


def get_device_list(daqmx_devices=None, verbose=False):
    """This function gets the list of Scope device in the system. If NI System
    Configuration is available, the list is grabbed from this library. Otherwise,
    the function attempts to use the `nilsdev` command line tool.

    Because `nilsdev` returns both DAQmx and Scope devices, the list of DAQmx devices
    is queried to remove them from the returned list. If the user code has already
    queried them, it is possible to pass them to avoid unnecessary double query.

    :param daqmx_devices: the list of DAQmx devices.
    :type daqmx_devices: list, optional
    :param verbose: sets verbosity level
    :type verbose: bool, optional
    """
    if has_nisyscfg:
        return {
            f"{devname:} (scope)": [f"{devname:}/{ii:d}" for ii in range(2)]
            for devname in scope_devices()
        }
    else:
        # NI-Scope backend
        # In principle, use nisyscfg, see essai_nisyscfg.py (does not work
        # on old Linux where NI System Configuration is too old.)
        # Workaround: call nilsdev. Any device not previously found by the
        # DAQmx backend is a NI-Scope (presumably)
        # (Only works on Linux, because there is no nilsdev on Windows)

        raw_data = sp.run("/usr/local/bin/nilsdev", capture_output=True)
        boards = dict()
        for line in raw_data.stdout.decode("ascii").split("\n"):
            line = line.strip()
            if "[Not Present]" in line:
                continue
            if not line:
                continue
            board_type, desc = line.split(":")
            desc = desc.strip()
            board_type = board_type.strip()
            if desc.startswith('"') and desc.endswith('"'):
                devname = desc[1:-1]
                boards[devname] = board_type
            else:
                print("Wrong format", desc)
        if daqmx_devices is None:
            if has_daqmx:
                daqmx_devices = daqmx_get_devices()
            else:
                print("Could not make sure boards are NI-Scope and not DAQmx")
                daqmx_devices = dict()
        for daqmx_board_desc, daqmx_devlist in daqmx_devices.items():
            board, devnum = daqmx_devlist[0].split("/")
            if verbose:
                print("DAQmx board:", board)
            if board in boards:
                del boards[board]
                if verbose:
                    print(f"Removing board {board:} from NI-Scope list.")
        return {
            f"{devname:} ({board_type:})": [f"{devname:}/{ii:d}" for ii in range(2)]
            for devname, board_type in boards.items()
        }



"""Asynchronous instruments module (:mod:`pymanip.aioinstruments`)
==================================================================

This module auto-imports all the asynchronous instrument classes.

"""

from .agilent_34970a import AsyncAgilent34970a
from .stanford_sr830 import AsyncStanfordSR830
from .newport_xps_rl import AsyncNewportXpsRL
from .tdk_lambda import AsyncTdkLambda
from .julabo import AsyncJulabo
from .tti_cpx400dp import AsyncTtiCpx400dp
from .lauda import AsyncLauda
from .arduino_gbf import AsyncArduino
